import json
import logging
import os
import uuid
from contextlib import contextmanager
from copy import deepcopy
from dataclasses import dataclass, fields
from errno import ENOENT
from functools import lru_cache
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

from recce.event import log_performance
from recce.exceptions import RecceException
from recce.util.cll import CLLPerformanceTracking, cll
from recce.util.lineage import (
    build_column_key,
    filter_dependency_maps,
    find_downstream,
    find_upstream,
)
from recce.util.perf_tracking import LineagePerfTracker

from ...tasks.profile import ProfileTask
from ...util.breaking import BreakingPerformanceTracking, parse_change_category

try:
    import agate
    import dbt.adapters.factory
    from dbt.contracts.state import PreviousState
except ImportError as e:
    print("Error: dbt module not found. Please install it by running:")
    print("pip install dbt-core dbt-<adapter>")
    raise e
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from recce.adapter.base import BaseAdapter
from recce.state import ArtifactsRoot

from ...models import RunType
from ...models.types import (
    CllColumn,
    CllData,
    CllNode,
    LineageDiff,
    NodeChange,
    NodeDiff,
)
from ...tasks import (
    HistogramDiffTask,
    ProfileDiffTask,
    QueryBaseTask,
    QueryDiffTask,
    QueryTask,
    RowCountDiffTask,
    RowCountTask,
    Task,
    TopKDiffTask,
    ValueDiffDetailTask,
    ValueDiffTask,
)
from .dbt_version import DbtVersion

dbt_supported_registry: Dict[RunType, Type[Task]] = {
    RunType.QUERY: QueryTask,
    RunType.QUERY_BASE: QueryBaseTask,
    RunType.QUERY_DIFF: QueryDiffTask,
    RunType.VALUE_DIFF: ValueDiffTask,
    RunType.VALUE_DIFF_DETAIL: ValueDiffDetailTask,
    RunType.PROFILE: ProfileTask,
    RunType.PROFILE_DIFF: ProfileDiffTask,
    RunType.ROW_COUNT: RowCountTask,
    RunType.ROW_COUNT_DIFF: RowCountDiffTask,
    RunType.TOP_K_DIFF: TopKDiffTask,
    RunType.HISTOGRAM_DIFF: HistogramDiffTask,
}

# Reference: https://github.com/AltimateAI/vscode-dbt-power-user/blob/master/dbt_core_integration.py

get_adapter_orig = dbt.adapters.factory.get_adapter


def get_adapter(config):
    if hasattr(config, "adapter"):
        return config.adapter
    else:
        return get_adapter_orig(config)


dbt.adapters.factory.get_adapter = get_adapter

# All dbt import should after overwriting the get_adapter
from dbt.adapters.base import Column  # noqa: E402
from dbt.adapters.factory import get_adapter_class_by_name  # noqa: E402
from dbt.adapters.sql import SQLAdapter  # noqa: E402
from dbt.config.runtime import RuntimeConfig  # noqa: E402
from dbt.contracts.graph.manifest import (  # noqa: E402
    MacroManifest,
    Manifest,
    WritableManifest,
)
from dbt.contracts.graph.nodes import ManifestNode  # noqa: E402
from dbt.contracts.results import CatalogArtifact  # noqa: E402
from dbt.flags import set_from_args  # noqa: E402
from dbt.parser.manifest import process_node  # noqa: E402
from dbt.parser.sql import SqlBlockParser  # noqa: E402

dbt_version = DbtVersion()

if dbt_version < "v1.8":
    from dbt.contracts.connection import Connection
else:
    from dbt.adapters.contracts.connection import Connection


@contextmanager
def silence_no_nodes_warning():
    if dbt_version >= "v1.8":
        from dbt.events.types import NoNodesForSelectionCriteria
        from dbt_common.events.functions import WARN_ERROR_OPTIONS

        WARN_ERROR_OPTIONS.silence.append(NoNodesForSelectionCriteria.__name__)
    try:
        yield
    finally:
        if dbt_version >= "v1.8":
            from dbt_common.events.functions import WARN_ERROR_OPTIONS

            WARN_ERROR_OPTIONS.silence.pop()


logger = logging.getLogger("uvicorn")
MIN_DBT_NODE_COMPOSITION = 3


class ArtifactsEventHandler(FileSystemEventHandler):
    def __init__(self, watch_files: List[str], callback: Callable = None):
        super().__init__()
        self.watch_files = watch_files
        self.callback = callback

    def on_modified(self, event):
        if event.is_directory:
            return None

        if event.src_path in self.watch_files:
            if callable(self.callback):
                self.callback(event)

    def on_created(self, event):
        if event.is_directory:
            return None

        if event.src_path in self.watch_files:
            if callable(self.callback):
                self.callback(event)


class EnvironmentEventHandler(FileSystemEventHandler):
    def __init__(self, observer, watch_files: set[str], callback: Callable = None):
        super().__init__()
        self.observer = observer
        self.watch_files = watch_files
        self.detected_files = set()
        self.callback = callback

    def on_created(self, event):
        if event.is_directory:
            return

        if event.src_path in self.watch_files:
            self.detected_files.add(event.src_path)

        # Check if all target-base files are created
        if self.detected_files == self.watch_files:
            if callable(self.callback):
                self.callback()
            self.observer.stop()


def merge_tables(tables: List[agate.Table]) -> agate.Table:
    if dbt_version < "v1.8":
        from dbt.clients.agate_helper import merge_tables

        return merge_tables(tables)
    else:
        from dbt_common.clients.agate_helper import merge_tables

        return merge_tables(tables)


def as_manifest(m: WritableManifest) -> Manifest:
    if dbt_version < "v1.8":
        data = m.__dict__
        all_fields = set([x.name for x in fields(Manifest)])
        new_data = {k: v for k, v in data.items() if k in all_fields}
        return Manifest(**new_data)
    else:
        return Manifest.from_writable_manifest(m)


def load_manifest(path: str = None, data: dict = None):
    if path is not None:
        if not os.path.isfile(path):
            return None
        return WritableManifest.read_and_check_versions(path)
    if data is not None:
        return WritableManifest.upgrade_schema_version(data)


def load_catalog(path: str = None, data: dict = None):
    if path is not None:
        if not os.path.isfile(path):
            return None
        return CatalogArtifact.read_and_check_versions(path)
    if data is not None:
        return CatalogArtifact.upgrade_schema_version(data)


def previous_state(state_path: Path, target_path: Path, project_root: Path) -> PreviousState:
    if dbt_version < "v1.5.2":
        return PreviousState(state_path, target_path)
    else:
        try:
            # Overwrite the level_tag method temporarily to avoid the warning message
            from dbt.events.types import EventLevel, WarnStateTargetEqual

            original_level_tag_func = WarnStateTargetEqual.level_tag
            WarnStateTargetEqual.level_tag = lambda x: EventLevel.DEBUG
        except ImportError:
            # Skip overwriting the level_tag method if the dbt version not support
            original_level_tag_func = None
            pass

        state = PreviousState(state_path, target_path, project_root)

        if original_level_tag_func is not None:
            # Restore the original level_tag method
            WarnStateTargetEqual.level_tag = original_level_tag_func

        return state


def default_profiles_dir():
    # Precedence: DBT_PROFILES_DIR > current working directory > ~/.dbt/
    # https://docs.getdbt.com/docs/core/connect-data-platform/connection-profiles#advanced-customizing-a-profile-directory
    if os.getenv("DBT_PROFILES_DIR"):
        return os.getenv("DBT_PROFILES_DIR")
    elif os.path.exists(os.path.join(os.getcwd(), "profiles.yml")):
        return os.getcwd()
    else:
        return os.path.expanduser("~/.dbt/")


@dataclass()
class DbtArgs:
    """
    Used for RuntimeConfig.from_args
    """

    threads: Optional[int] = (1,)
    target: Optional[str] = (None,)
    profiles_dir: Optional[str] = (None,)
    project_dir: Optional[str] = (None,)
    profile: Optional[str] = (None,)
    target_path: Optional[str] = (None,)
    project_only_flags: Optional[Dict[str, Any]] = None
    which: Optional[str] = None
    state_modified_compare_more_unrendered_values: Optional[bool] = True  # new flag added since dbt v1.9


@dataclass
class DbtAdapter(BaseAdapter):
    runtime_config: RuntimeConfig = None
    adapter: SQLAdapter = None
    manifest: Manifest = None
    previous_state: PreviousState = None
    target_path: str = None
    curr_manifest: WritableManifest = None
    curr_catalog: CatalogArtifact = None
    base_path: str = None
    base_manifest: WritableManifest = None
    base_catalog: CatalogArtifact = None

    # Review mode
    review_mode: bool = False

    # Watch the artifact change
    artifacts_observer = Observer()
    artifacts_files = []

    # Watch the base environment ready
    base_env_observer = Observer()

    def support_tasks(self):
        support_map = {run_type.value: True for run_type in dbt_supported_registry}

        return support_map

    @classmethod
    def load(cls, no_artifacts=False, review=False, **kwargs):
        target = kwargs.get("target")
        target_path = kwargs.get("target_path", "target")
        target_base_path = kwargs.get("target_base_path", "target-base")

        profile_name = kwargs.get("profile")
        project_dir = kwargs.get("project_dir")
        profiles_dir = kwargs.get("profiles_dir")

        if profiles_dir is None:
            profiles_dir = default_profiles_dir()

        # runtime_config
        args = DbtArgs(
            threads=1,
            target=target,
            target_path=target_path,
            project_dir=project_dir,
            profiles_dir=profiles_dir,
            profile=profile_name,
            project_only_flags={},
            which="list",
        )
        set_from_args(args, args)

        from dbt.exceptions import DbtProjectError

        try:
            # adapter
            if dbt_version < "v1.8":
                runtime_config = RuntimeConfig.from_args(args)
                adapter_name = runtime_config.credentials.type
                adapter_cls = get_adapter_class_by_name(adapter_name)
                adapter: SQLAdapter = adapter_cls(runtime_config)
            else:
                from dbt.mp_context import get_mp_context
                from dbt_common.context import (
                    get_invocation_context,
                    set_invocation_context,
                )

                set_invocation_context({})
                get_invocation_context()._env = dict(os.environ)
                runtime_config = RuntimeConfig.from_args(args)
                adapter_name = runtime_config.credentials.type
                adapter_cls = get_adapter_class_by_name(adapter_name)
                adapter: SQLAdapter = adapter_cls(runtime_config, get_mp_context())
                from dbt.adapters.factory import FACTORY

                FACTORY.adapters[adapter_name] = adapter

            adapter.connections.set_connection_name()
            runtime_config.adapter = adapter

            dbt_adapter = cls(
                runtime_config=runtime_config,
                adapter=adapter,
                review_mode=review,
                base_path=target_base_path,
            )
        except DbtProjectError as e:
            raise e

        # Load the artifacts from the state file or dbt target and dbt base directory
        if not no_artifacts and not review:
            dbt_adapter.load_artifacts()
        return dbt_adapter

    def print_lineage_info(self):
        print("Base:")
        print(f"    Manifest: {self.base_manifest.metadata.generated_at}")
        print(f"    Catalog:  {self.base_catalog.metadata.generated_at if self.base_catalog else 'N/A'}")

        print("Current:")
        print(f"    Manifest: {self.curr_manifest.metadata.generated_at}")
        print(f"    Catalog:  {self.curr_catalog.metadata.generated_at if self.curr_catalog else 'N/A'}")

    def get_columns(self, model: str, base=False) -> List[Column]:
        relation = self.create_relation(model, base)
        get_columns_macro = "get_columns_in_relation"
        if self.adapter.connections.TYPE == "databricks":
            get_columns_macro = "get_columns_comments"

        if dbt_version < "v1.8":
            columns = self.adapter.execute_macro(
                get_columns_macro, kwargs={"relation": relation}, manifest=self.manifest
            )
        else:
            from dbt.context.providers import generate_runtime_macro_context

            macro_manifest = MacroManifest(self.manifest.macros)
            self.adapter.set_macro_resolver(macro_manifest)
            self.adapter.set_macro_context_generator(generate_runtime_macro_context)
            columns = self.adapter.execute_macro(get_columns_macro, kwargs={"relation": relation})

        if self.adapter.connections.TYPE == "databricks":
            # reference: get_columns_in_relation (dbt/adapters/databricks/impl.py)
            from dbt.adapters.databricks.column import DatabricksColumn

            rows = columns
            columns = []
            for row in rows:
                if row["col_name"].startswith("#"):
                    break
                columns.append(
                    DatabricksColumn(
                        column=row["col_name"],
                        dtype=row["data_type"],
                        comment=row["comment"],
                    )
                )
            return columns
        else:
            return columns

    def get_model(self, model_id: str, base=False):
        manifest = self.curr_manifest if base is False else self.base_manifest
        manifest_dict = manifest.to_dict()

        node = manifest_dict["nodes"].get(model_id)
        if node is None:
            return {}

        node_name = node["name"]
        with self.adapter.connection_named("model"):
            columns = [column for column in self.get_columns(node_name, base=base)]

        child_map: List[str] = manifest_dict["child_map"][model_id]
        cols_not_null = []
        cols_unique = []

        for child in child_map:
            comps = child.split(".")
            child_type = comps[0]
            child_name = comps[2]

            not_null_prefix = f"not_null_{node_name}_"
            if child_type == "test" and child_name.startswith(not_null_prefix):
                cols_not_null.append(child_name[len(not_null_prefix) :])
            unique_prefix = f"unique_{node_name}_"
            if child_type == "test" and child_name.startswith(unique_prefix):
                cols_unique.append(child_name[len(unique_prefix) :])

        columns_info = {}
        primary_key = None
        for c in columns:
            col_name = c.column
            col = dict(name=col_name, type=c.dtype)
            if col_name in cols_not_null:
                col["not_null"] = True
            if col_name in cols_unique:
                col["unique"] = True
                if not primary_key:
                    primary_key = col_name
            columns_info[col_name] = col

        result = dict(columns=columns_info)
        if primary_key:
            result["primary_key"] = primary_key

        return result

    def load_artifacts(self):
        """
        Load the artifacts from the 'target' and 'target-base' directory
        """
        if self.runtime_config is None:
            raise Exception("Cannot find the dbt project configuration")

        project_root = self.runtime_config.project_root
        target_path = self.runtime_config.target_path
        target_base_path = self.base_path
        self.target_path = os.path.join(project_root, target_path)
        self.base_path = os.path.join(project_root, target_base_path)

        # load the artifacts
        path = os.path.join(project_root, target_path, "manifest.json")
        curr_manifest = load_manifest(path=path)
        if curr_manifest is None:
            raise FileNotFoundError(ENOENT, os.strerror(ENOENT), path)
        path = os.path.join(project_root, target_base_path, "manifest.json")
        base_manifest = load_manifest(path=path)
        if base_manifest is None:
            raise FileNotFoundError(ENOENT, os.strerror(ENOENT), path)

        curr_catalog = load_catalog(path=os.path.join(project_root, target_path, "catalog.json"))
        base_catalog = load_catalog(path=os.path.join(project_root, target_base_path, "catalog.json"))

        # set the value if all the artifacts are loaded successfully
        self.curr_manifest = curr_manifest
        self.curr_catalog = curr_catalog
        self.base_manifest = base_manifest
        self.base_catalog = base_catalog

        # set the manifest
        self.manifest = as_manifest(curr_manifest)
        self.previous_state = previous_state(
            Path(target_base_path),
            Path(self.runtime_config.target_path),
            Path(self.runtime_config.project_root),
        )

        # set the file paths to watch
        self.artifacts_files = [
            os.path.join(project_root, target_path, "manifest.json"),
            os.path.join(project_root, target_path, "catalog.json"),
            os.path.join(project_root, target_base_path, "manifest.json"),
            os.path.join(project_root, target_base_path, "catalog.json"),
        ]

    def is_python_model(self, node_id: str, base: Optional[bool] = False):
        manifest = self.curr_manifest if base is False else self.base_manifest
        model = manifest.nodes.get(node_id)
        if hasattr(model, "language"):
            return model.language == "python"

        return False

    def find_node_by_name(self, node_name, base=False) -> Optional[ManifestNode]:
        manifest = self.curr_manifest if base is False else self.base_manifest

        for key, node in manifest.nodes.items():
            if node.name == node_name:
                return node

        return None

    def get_node_name_by_id(self, unique_id):
        if unique_id.startswith("source."):
            if unique_id in self.curr_manifest.sources:
                return self.curr_manifest.sources[unique_id].name
            elif unique_id in self.base_manifest.sources:
                return self.base_manifest.sources[unique_id].name
        elif unique_id.startswith("metric."):
            if unique_id in self.curr_manifest.metrics:
                return self.curr_manifest.metrics[unique_id].name
            elif unique_id in self.base_manifest.metrics:
                return self.base_manifest.metrics[unique_id].name
        elif unique_id.startswith("exposure."):
            if unique_id in self.curr_manifest.exposures:
                return self.curr_manifest.exposures[unique_id].name
            elif unique_id in self.base_manifest.exposures:
                return self.base_manifest.exposures[unique_id].name
        elif unique_id.startswith("semantic_model."):
            if unique_id in self.curr_manifest.semantic_models:
                return self.curr_manifest.semantic_models[unique_id].name
            elif unique_id in self.base_manifest.semantic_models:
                return self.base_manifest.semantic_models[unique_id].name
        else:
            if unique_id in self.curr_manifest.nodes:
                return self.curr_manifest.nodes[unique_id].name
            elif unique_id in self.base_manifest.nodes:
                return self.base_manifest.nodes[unique_id].name
            return None

    def get_manifest(self, base: bool):
        return self.curr_manifest if base is False else self.base_manifest

    def generate_sql(
        self,
        sql_template: str,
        base: bool = False,
        context=None,
        provided_manifest=None,
    ):
        if context is None:
            context = {}
        manifest = provided_manifest if provided_manifest is not None else as_manifest(self.get_manifest(base))
        parser = SqlBlockParser(self.runtime_config, manifest, self.runtime_config)

        if dbt_version >= dbt_version.parse("v1.8"):
            from dbt_common.context import (
                get_invocation_context,
                set_invocation_context,
            )

            set_invocation_context({})
            get_invocation_context()._env = dict(os.environ)

        node_id = str("generated_" + uuid.uuid4().hex)
        node = parser.parse_remote(sql_template, node_id)
        process_node(self.runtime_config, manifest, node)

        if dbt_version < dbt_version.parse("v1.8"):
            compiler = self.adapter.get_compiler()
            compiler.compile_node(node, manifest, context)
            return node.compiled_code
        else:
            from dbt.clients import jinja
            from dbt.context.providers import (
                generate_runtime_macro_context,
                generate_runtime_model_context,
            )

            # Set up macro resolver for dbt >= 1.8
            macro_manifest = MacroManifest(manifest.macros)
            self.adapter.set_macro_resolver(macro_manifest)
            self.adapter.set_macro_context_generator(generate_runtime_macro_context)

            jinja_ctx = generate_runtime_model_context(node, self.runtime_config, manifest)
            jinja_ctx.update(context)
            compiled_code = jinja.get_rendered(sql_template, jinja_ctx, node)
            return compiled_code

    def execute(
        self,
        sql: str,
        auto_begin: bool = False,
        fetch: bool = False,
        limit: Optional[int] = None,
    ) -> Tuple[any, agate.Table]:
        if dbt_version < dbt_version.parse("v1.6"):
            return self.adapter.execute(sql, auto_begin=auto_begin, fetch=fetch)

        return self.adapter.execute(sql, auto_begin=auto_begin, fetch=fetch, limit=limit)

    def build_parent_map(self, nodes: Dict, base: Optional[bool] = False) -> Dict[str, List[str]]:
        manifest = self.curr_manifest if base is False else self.base_manifest
        manifest_dict = manifest.to_dict()

        node_ids = nodes.keys()
        parent_map = {}
        for k, parents in manifest_dict["parent_map"].items():
            if k not in node_ids:
                continue
            parent_map[k] = [parent for parent in parents if parent in node_ids]

        return parent_map

    def build_parent_list_per_node(self, node_id: str, base: Optional[bool] = False) -> List[str]:
        manifest = self.curr_manifest if base is False else self.base_manifest
        manifest_dict = manifest.to_dict()

        if node_id in manifest_dict["parent_map"]:
            return manifest_dict["parent_map"][node_id]

    def get_lineage(self, base: Optional[bool] = False):
        manifest = self.curr_manifest if base is False else self.base_manifest
        catalog = self.curr_catalog if base is False else self.base_catalog
        cache_key = hash((id(manifest), id(catalog)))
        return self.get_lineage_cached(base, cache_key)

    def get_lineage_diff(self) -> LineageDiff:
        cache_key = hash(
            (
                id(self.base_manifest),
                id(self.base_catalog),
                id(self.curr_manifest),
                id(self.curr_catalog),
            )
        )
        return self._get_lineage_diff_cached(cache_key)

    @lru_cache(maxsize=2)
    def get_lineage_cached(self, base: Optional[bool] = False, cache_key=0):
        if base is False:
            perf_tracker = LineagePerfTracker()
            perf_tracker.start_lineage()

        manifest = self.curr_manifest if base is False else self.base_manifest
        catalog = self.curr_catalog if base is False else self.base_catalog

        manifest_metadata = manifest.metadata if manifest is not None else None
        catalog_metadata = catalog.metadata if catalog is not None else None

        manifest_dict = manifest.to_dict()

        nodes = {}

        for node in manifest_dict["nodes"].values():
            unique_id = node["unique_id"]
            resource_type = node["resource_type"]

            if resource_type not in ["model", "seed", "exposure", "snapshot"]:
                continue

            nodes[unique_id] = {
                "id": node["unique_id"],
                "name": node["name"],
                "resource_type": node["resource_type"],
                "package_name": node["package_name"],
                "schema": node["schema"],
                "config": node["config"],
                "checksum": node["checksum"],
                "raw_code": node["raw_code"],
            }

            # List of <type>.<package_name>.<node_name>.<hash>
            # model.jaffle_shop.customer_segments
            # test.jaffle_shop.not_null_customers_customer_id.5c9bf9911d
            # test.jaffle_shop.unique_customers_customer_id.c5af1ff4b1
            child_map: List[str] = manifest_dict["child_map"][unique_id]
            cols_not_null = []
            cols_unique = []

            for child in child_map:
                node_name = node["name"]
                comps = child.split(".")
                if len(comps) < MIN_DBT_NODE_COMPOSITION:
                    # only happens in unittest
                    continue

                child_type = comps[0]
                child_name = comps[2]

                not_null_prefix = f"not_null_{node_name}_"
                if child_type == "test" and child_name.startswith(not_null_prefix):
                    cols_not_null.append(child_name[len(not_null_prefix) :])
                unique_prefix = f"unique_{node_name}_"
                if child_type == "test" and child_name.startswith(unique_prefix):
                    cols_unique.append(child_name[len(unique_prefix) :])

            if catalog is not None and unique_id in catalog.nodes:
                columns = {}
                primary_key = None
                for col_name, col_metadata in catalog.nodes[unique_id].columns.items():
                    col = dict(name=col_name, type=col_metadata.type)
                    if col_name in cols_not_null:
                        col["not_null"] = True
                    if col_name in cols_unique:
                        col["unique"] = True
                        if not primary_key:
                            primary_key = col_name
                    columns[col_name] = col
                nodes[unique_id]["columns"] = columns
                if primary_key:
                    nodes[unique_id]["primary_key"] = primary_key

        for source in manifest_dict["sources"].values():
            unique_id = source["unique_id"]

            nodes[unique_id] = {
                "id": source["unique_id"],
                "name": source["name"],
                "resource_type": source["resource_type"],
                "package_name": source["package_name"],
                "config": source["config"],
            }

            if catalog is not None and unique_id in catalog.sources:
                nodes[unique_id]["columns"] = {
                    col_name: {"name": col_name, "type": col_metadata.type}
                    for col_name, col_metadata in catalog.sources[unique_id].columns.items()
                }

        for exposure in manifest_dict["exposures"].values():
            nodes[exposure["unique_id"]] = {
                "id": exposure["unique_id"],
                "name": exposure["name"],
                "resource_type": exposure["resource_type"],
                "package_name": exposure["package_name"],
                "config": exposure["config"],
            }
        for metric in manifest_dict["metrics"].values():
            nodes[metric["unique_id"]] = {
                "id": metric["unique_id"],
                "name": metric["name"],
                "resource_type": metric["resource_type"],
                "package_name": metric["package_name"],
                "config": metric["config"],
            }

        if "semantic_models" in manifest_dict:
            for semantic_models in manifest_dict["semantic_models"].values():
                nodes[semantic_models["unique_id"]] = {
                    "id": semantic_models["unique_id"],
                    "name": semantic_models["name"],
                    "resource_type": semantic_models["resource_type"],
                    "package_name": semantic_models["package_name"],
                    "config": semantic_models["config"],
                }

        parent_map = self.build_parent_map(nodes, base)

        if base is False:
            perf_tracker.end_lineage()
            perf_tracker.set_total_nodes(len(nodes))
            log_performance("model lineage", perf_tracker.to_dict())
            perf_tracker.reset()

        return dict(
            parent_map=parent_map,
            nodes=nodes,
            manifest_metadata=manifest_metadata,
            catalog_metadata=catalog_metadata,
        )

    @lru_cache(maxsize=1)
    def _get_lineage_diff_cached(self, cache_key) -> LineageDiff:
        base = self.get_lineage(base=True)
        current = self.get_lineage(base=False)

        modified_nodes = self.select_nodes(select="state:modified")
        diff = {}
        for node_id in modified_nodes:
            base_node = base.get("nodes", {}).get(node_id)
            curr_node = current.get("nodes", {}).get(node_id)
            if base_node and curr_node:
                diff[node_id] = NodeDiff(change_status="modified")
            elif base_node:
                diff[node_id] = NodeDiff(change_status="removed")
            elif curr_node:
                diff[node_id] = NodeDiff(change_status="added")

        return LineageDiff(
            base=base,
            current=current,
            diff=diff,
        )

    @lru_cache(maxsize=128)
    def get_change_analysis_cached(self, node_id: str):
        breaking_perf_tracker = BreakingPerformanceTracking()
        lineage_diff = self.get_lineage_diff()
        diff = lineage_diff.diff

        if node_id not in diff or diff[node_id].change_status != "modified":
            return diff.get(node_id)

        breaking_perf_tracker.increment_modified_nodes()
        breaking_perf_tracker.start_lineage_diff()

        base = lineage_diff.base
        current = lineage_diff.current

        base_manifest = as_manifest(self.get_manifest(True))
        curr_manifest = as_manifest(self.get_manifest(False))
        breaking_perf_tracker.record_checkpoint("manifest")

        def ref_func(*args):
            if len(args) == 1:
                node = args[0]
            elif len(args) > 1:
                node = args[1]
            else:
                return None
            return node

        def source_func(source_name, table_name):
            source_name = source_name.replace("-", "_")
            return f"__{source_name}__{table_name}"

        jinja_context = dict(
            ref=ref_func,
            source=source_func,
        )

        base_node = base.get("nodes", {}).get(node_id)
        curr_node = current.get("nodes", {}).get(node_id)
        change = NodeChange(category="unknown")
        if (
            curr_node.get("resource_type") in ["model", "snapshot"]
            and curr_node.get("raw_code") is not None
            and base_node.get("raw_code") is not None
        ):
            try:

                def _get_schema(lineage):
                    schema = {}
                    nodes = lineage["nodes"]
                    parent_list = lineage["parent_map"].get(node_id, [])
                    for parent_id in parent_list:
                        parent_node = nodes.get(parent_id)
                        if parent_node is None:
                            continue
                        columns = parent_node.get("columns") or {}
                        name = parent_node.get("name")
                        if parent_node.get("resource_type") == "source":
                            parts = parent_id.split(".")
                            source = parts[2]
                            table = parts[3]
                            source = source.replace("-", "_")
                            name = f"__{source}__{table}"
                        schema[name] = {name: column.get("type") for name, column in columns.items()}
                    return schema

                base_sql = self.generate_sql(
                    base_node.get("raw_code"),
                    context=jinja_context,
                    provided_manifest=base_manifest,
                )
                curr_sql = self.generate_sql(
                    curr_node.get("raw_code"),
                    context=jinja_context,
                    provided_manifest=curr_manifest,
                )
                base_schema = _get_schema(base)
                curr_schema = _get_schema(current)
                dialect = self.adapter.connections.TYPE
                if curr_manifest.metadata.adapter_type is not None:
                    dialect = curr_manifest.metadata.adapter_type

                change = parse_change_category(
                    base_sql,
                    curr_sql,
                    old_schema=base_schema,
                    new_schema=curr_schema,
                    dialect=dialect,
                    perf_tracking=breaking_perf_tracker,
                )

                # Make sure that the case of the column names are the same
                changed_columns = {
                    column.lower(): change_status for column, change_status in (change.columns or {}).items()
                }
                changed_columns_names = set(changed_columns)
                changed_columns_final = {}

                base_columns = base_node.get("columns") or {}
                curr_columns = curr_node.get("columns") or {}
                columns_names = set(base_columns) | set(curr_columns)

                for column_name in columns_names:
                    if column_name.lower() in changed_columns_names:
                        changed_columns_final[column_name] = changed_columns[column_name.lower()]

                change.columns = changed_columns_final
            except Exception:
                # TODO: telemetry
                pass

        breaking_perf_tracker.end_lineage_diff()
        log_performance("change analysis per node", breaking_perf_tracker.to_dict())
        breaking_perf_tracker.reset()
        node_diff = diff.get(node_id)
        node_diff.change = change
        return node_diff

    def get_cll(
        self,
        node_id: Optional[str] = None,
        column: Optional[str] = None,
        change_analysis: Optional[bool] = False,
        no_cll: Optional[bool] = False,
        no_upstream: Optional[bool] = False,
        no_downstream: Optional[bool] = False,
        no_filter: Optional[bool] = False,
    ) -> CllData:
        cll_tracker = LineagePerfTracker()
        cll_tracker.set_params(
            has_node=node_id is not None,
            has_column=column is not None,
            change_analysis=change_analysis,
            no_cll=no_cll,
            no_upstream=no_upstream,
            no_downstream=no_downstream,
        )
        cll_tracker.start_column_lineage()

        manifest = self.curr_manifest
        manifest_dict = manifest.to_dict()

        # Find related model nodes
        if node_id is not None:
            cll_node_ids = {node_id}
        else:
            lineage_diff = self.get_lineage_diff()
            cll_node_ids = set(lineage_diff.diff.keys())

        cll_tracker.set_init_nodes(len(cll_node_ids))

        nodes = {}
        columns = {}
        parent_map = {}
        child_map = {}

        if not no_upstream:
            cll_node_ids = cll_node_ids.union(find_upstream(cll_node_ids, manifest_dict.get("parent_map")))
        if not no_downstream:
            cll_node_ids = cll_node_ids.union(find_downstream(cll_node_ids, manifest_dict.get("child_map")))

        if not no_cll:
            allowed_related_nodes = set()
            for key in ["sources", "nodes", "exposures", "metrics"]:
                attr = getattr(manifest, key)
                allowed_related_nodes.update(set(attr.keys()))
            if hasattr(manifest, "semantic_models"):
                attr = getattr(manifest, "semantic_models")
                allowed_related_nodes.update(set(attr.keys()))
            for cll_node_id in cll_node_ids:
                if cll_node_id not in allowed_related_nodes:
                    continue
                cll_data_one = deepcopy(self.get_cll_cached(cll_node_id, base=False))
                cll_tracker.increment_cll_nodes()
                if cll_data_one is None:
                    continue

                nodes[cll_node_id] = cll_data_one.nodes.get(cll_node_id)
                node_diff = None
                if change_analysis:
                    node_diff = self.get_change_analysis_cached(cll_node_id)
                    cll_tracker.increment_change_analysis_nodes()
                if node_diff is not None:
                    nodes[cll_node_id].change_status = node_diff.change_status
                    if node_diff.change is not None:
                        nodes[cll_node_id].change_category = node_diff.change.category
                for c_id, c in cll_data_one.columns.items():
                    columns[c_id] = c
                    if node_diff is not None:
                        if node_diff.change_status == "added":
                            c.change_status = "added"
                        elif node_diff.change_status == "removed":
                            c.change_status = "removed"
                        elif node_diff.change is not None and node_diff.change.columns is not None:
                            column_diff = node_diff.change.columns.get(c.name)
                            if column_diff:
                                c.change_status = column_diff

                for p_id, parents in cll_data_one.parent_map.items():
                    parent_map[p_id] = parents
        else:
            for cll_node_id in cll_node_ids:
                cll_node = None
                cll_node_columns: Dict[str, CllColumn] = {}

                if cll_node_id in manifest.sources:
                    cll_node = CllNode.build_cll_node(manifest, "sources", cll_node_id)
                    if self.curr_catalog and cll_node_id in self.curr_catalog.sources:
                        cll_node_columns = {
                            column.name: CllColumn(
                                id=f"{cll_node_id}_{column.name}",
                                table_id=cll_node_id,
                                name=column.name,
                                type=column.type,
                            )
                            for column in self.curr_catalog.sources[cll_node_id].columns.values()
                        }
                elif cll_node_id in manifest.nodes:
                    cll_node = CllNode.build_cll_node(manifest, "nodes", cll_node_id)
                    if self.curr_catalog and cll_node_id in self.curr_catalog.nodes:
                        cll_node_columns = {
                            column.name: CllColumn(
                                id=f"{cll_node_id}_{column.name}",
                                table_id=cll_node_id,
                                name=column.name,
                                type=column.type,
                            )
                            for column in self.curr_catalog.nodes[cll_node_id].columns.values()
                        }
                elif cll_node_id in manifest.exposures:
                    cll_node = CllNode.build_cll_node(manifest, "exposures", cll_node_id)
                elif hasattr(manifest, "semantic_models") and cll_node_id in manifest.semantic_models:
                    cll_node = CllNode.build_cll_node(manifest, "semantic_models", cll_node_id)
                elif cll_node_id in manifest.metrics:
                    cll_node = CllNode.build_cll_node(manifest, "metrics", cll_node_id)

                if not cll_node:
                    continue
                nodes[cll_node_id] = cll_node

                node_diff = None
                if change_analysis:
                    node_diff = self.get_change_analysis_cached(cll_node_id)
                    cll_tracker.increment_change_analysis_nodes()
                if node_diff is not None:
                    cll_node.change_status = node_diff.change_status
                    if node_diff.change is not None:
                        cll_node.change_category = node_diff.change.category
                        for c, cll_column in cll_node_columns.items():
                            cll_node.columns[c] = cll_column
                            columns[cll_column.id] = cll_column
                            if node_diff.change.columns and c in node_diff.change.columns:
                                cll_column.change_status = node_diff.change.columns[c]

                parent_map[cll_node_id] = manifest.parent_map.get(cll_node_id, [])

        # build the child map
        for parent_id, parents in parent_map.items():
            for parent in parents:
                if parent not in child_map:
                    child_map[parent] = set()
                child_map[parent].add(parent_id)

        # Find the anchor nodes
        anchor_node_ids = set()
        extra_node_ids = set()
        if node_id is None and column is None:
            if change_analysis:
                # If change analysis is requested, we need to find the nodes that have changes
                lineage_diff = self.get_lineage_diff()
                for nid, nd in lineage_diff.diff.items():
                    if nd.change_status == "added":
                        anchor_node_ids.add(nid)
                        n = lineage_diff.current["nodes"].get(nid)
                        n_columns = n.get("columns", {})
                        for c in n_columns:
                            anchor_node_ids.add(build_column_key(nid, c))
                        continue
                    if nd.change_status == "removed":
                        extra_node_ids.add(nid)
                        continue

                    node_diff = self.get_change_analysis_cached(nid)
                    if node_diff is not None and node_diff.change is not None:
                        extra_node_ids.add(nid)
                        if no_cll:
                            if node_diff.change.category in ["breaking", "partial_breaking", "unknown"]:
                                anchor_node_ids.add(nid)
                        else:
                            if node_diff.change.category in ["breaking", "unknown"]:
                                anchor_node_ids.add(nid)
                        if node_diff.change.columns is not None:
                            for column_name in node_diff.change.columns:
                                anchor_node_ids.add(f"{nid}_{column_name}")
            else:
                lineage_diff = self.get_lineage_diff()
                anchor_node_ids = lineage_diff.diff.keys()
        elif node_id is not None and column is None:
            if change_analysis:
                # If change analysis is requested, we need to find the nodes that have changes
                node_diff = self.get_change_analysis_cached(node_id)
                if node_diff is not None and node_diff.change is not None:
                    extra_node_ids.add(node_id)
                    if no_cll:
                        if node_diff.change.category in ["breaking", "partial_breaking", "unknown"]:
                            anchor_node_ids.add(node_id)
                    else:
                        if node_diff.change.category in ["breaking", "unknown"]:
                            anchor_node_ids.add(node_id)
                    if node_diff.change.columns is not None:
                        for column_name in node_diff.change.columns:
                            anchor_node_ids.add(f"{node_id}_{column_name}")
                else:
                    anchor_node_ids.add(node_id)
            else:
                anchor_node_ids.add(node_id)
                if not no_cll:
                    node = nodes.get(node_id)
                    if node:
                        for column_name in node.columns:
                            column_key = build_column_key(node_id, column_name)
                            anchor_node_ids.add(column_key)
        else:
            anchor_node_ids.add(f"{node_id}_{column}")

        cll_tracker.set_anchor_nodes(len(anchor_node_ids))
        result_node_ids = set(anchor_node_ids)
        if not no_upstream:
            result_node_ids = result_node_ids.union(find_upstream(anchor_node_ids, parent_map))
        if not no_downstream:
            result_node_ids = result_node_ids.union(find_downstream(anchor_node_ids, child_map))

        # Filter the nodes and columns based on the anchor nodes
        if not no_filter:
            nodes = {k: v for k, v in nodes.items() if k in result_node_ids or k in extra_node_ids}
            columns = {k: v for k, v in columns.items() if k in result_node_ids or k in extra_node_ids}

            for node in nodes.values():
                node.columns = {
                    k: v for k, v in node.columns.items() if v.id in result_node_ids or v.id in extra_node_ids
                }

                if change_analysis:
                    node.impacted = node.id in result_node_ids

            parent_map, child_map = filter_dependency_maps(parent_map, child_map, result_node_ids)

        cll_tracker.end_column_lineage()
        cll_tracker.set_total_nodes(len(nodes) + len(columns))
        log_performance("column level lineage", cll_tracker.to_dict())
        cll_tracker.reset()

        return CllData(
            nodes=nodes,
            columns=columns,
            parent_map=parent_map,
            child_map=child_map,
        )

    @lru_cache(maxsize=128)
    def get_cll_cached(self, node_id: str, base: Optional[bool] = False) -> Optional[CllData]:
        cll_tracker = CLLPerformanceTracking()

        node, parent_list = self.get_cll_node(node_id, base=base)
        if node is None:
            return None

        cll_tracker.set_total_nodes(1)
        cll_tracker.start_column_lineage()

        def _apply_all_columns(node: CllNode, transformation_type):
            cll_data = CllData()
            cll_data.nodes[node.id] = node
            cll_data.parent_map[node.id] = set(parent_list)
            for col in node.columns.values():
                column_id = f"{node.id}_{col.name}"
                col.transformation_type = transformation_type
                cll_data.columns[column_id] = col
                cll_data.parent_map[column_id] = set()
            return cll_data

        manifest = as_manifest(self.get_manifest(base))
        catalog = self.curr_catalog if base is False else self.base_catalog
        resource_type = node.resource_type
        if resource_type not in {"model", "seed", "source", "snapshot"}:
            return _apply_all_columns(node, "unknown")

        if resource_type == "source" or resource_type == "seed":
            return _apply_all_columns(node, "source")

        if node.raw_code is None or self.is_python_model(node.id, base=base):
            return _apply_all_columns(node, "unknown")

        if node.name == "metricflow_time_spine":
            return _apply_all_columns(node, "source")

        if not node.columns:
            return _apply_all_columns(node, "unknown")

        table_id_map = {}

        def ref_func(*args):
            node_name: str = None
            project_or_package: str = None

            if len(args) == 1:
                node_name = args[0]
            else:
                project_or_package = args[0]
                node_name = args[1]

            for key, n in manifest.nodes.items():
                if n.name != node_name:
                    continue
                if project_or_package is not None and n.package_name != project_or_package:
                    continue

                # replace id "." to "_"
                unique_id = n.unique_id
                table_name = unique_id.replace(".", "_")
                table_id_map[table_name.lower()] = unique_id
                return table_name

            raise ValueError(f"Cannot find node {node_name} in the manifest")

        def source_func(source_name, name):
            for key, n in manifest.sources.items():
                if n.source_name != source_name:
                    continue
                if n.name != name:
                    continue

                # replace id "." to "_"
                unique_id = n.unique_id
                table_name = unique_id.replace(".", "_")
                table_id_map[table_name.lower()] = unique_id
                return table_name

            raise ValueError(f"Cannot find source {source_name}.{name} in the manifest")

        raw_code = node.raw_code
        jinja_context = dict(
            ref=ref_func,
            source=source_func,
        )

        schema = {}
        if catalog is not None:
            for parent_id in parent_list:
                table_name = parent_id.replace(".", "_")
                columns = {}
                if parent_id in catalog.nodes:
                    for col_name, col_metadata in catalog.nodes[parent_id].columns.items():
                        columns[col_name] = col_metadata.type
                if parent_id in catalog.sources:
                    for col_name, col_metadata in catalog.sources[parent_id].columns.items():
                        columns[col_name] = col_metadata.type
                schema[table_name] = columns

        try:
            compiled_sql = self.generate_sql(raw_code, base=base, context=jinja_context, provided_manifest=manifest)
            dialect = self.adapter.type()
            if self.get_manifest(base).metadata.adapter_type is not None:
                dialect = self.get_manifest(base).metadata.adapter_type
            m2c, c2c_map = cll(compiled_sql, schema=schema, dialect=dialect)
        except RecceException:
            cll_tracker.increment_sqlglot_error_nodes()
            return _apply_all_columns(node, "unknown")
        except Exception:
            cll_tracker.increment_other_error_nodes()
            return _apply_all_columns(node, "unknown")

        # Add cll dependency to the node.
        cll_data = CllData()
        cll_data.nodes[node.id] = node
        cll_data.columns = {f"{node.id}_{col.name}": col for col in node.columns.values()}

        # parent map for node
        depends_on = set(parent_list)
        for d in m2c:
            parent_key = f"{table_id_map[d.node.lower()]}_{d.column}"
            depends_on.add(parent_key)
        cll_data.parent_map[node_id] = depends_on

        # parent map for columns
        for name, column in node.columns.items():
            depends_on = set()
            column_id = f"{node.id}_{name}"
            if name in c2c_map:
                for d in c2c_map[name].depends_on:
                    parent_key = f"{table_id_map[d.node.lower()]}_{d.column}"
                    depends_on.add(parent_key)
                column.transformation_type = c2c_map[name].transformation_type
            cll_data.parent_map[column_id] = set(depends_on)

        cll_tracker.end_column_lineage()
        log_performance("column level lineage per node", cll_tracker.to_dict())
        cll_tracker.reset()
        return cll_data

    def get_cll_node(self, node_id: str, base: Optional[bool] = False) -> Tuple[Optional[CllNode], list[str]]:
        manifest = self.curr_manifest if base is False else self.base_manifest
        catalog = self.curr_catalog if base is False else self.base_catalog
        parent_list = []
        node = None

        # model, seed, snapshot
        if node_id in manifest.nodes:
            found = manifest.nodes[node_id]
            unique_id = found.unique_id
            node = CllNode.build_cll_node(manifest, "nodes", node_id)
            if hasattr(found.depends_on, "nodes"):
                parent_list = found.depends_on.nodes

            if catalog is not None and node is not None and unique_id in catalog.nodes:
                columns = {}
                for col_name, col_metadata in catalog.nodes[unique_id].columns.items():
                    column_id = f"{unique_id}_{col_name}"
                    col = CllColumn(id=column_id, name=col_name, table_id=unique_id, type=col_metadata.type)
                    columns[col_name] = col
                node.columns = columns

        # source
        if node_id in manifest.sources:
            found = manifest.sources[node_id]
            unique_id = found.unique_id
            node = CllNode.build_cll_node(manifest, "sources", node_id)
            parent_list = []

            if catalog is not None and node is not None and unique_id in catalog.sources:
                columns = {}
                for col_name, col_metadata in catalog.sources[unique_id].columns.items():
                    column_id = f"{unique_id}_{col_name}"
                    col = CllColumn(id=column_id, name=col_name, table_id=unique_id, type=col_metadata.type)
                    columns[col_name] = col
                node.columns = columns

        # exposure
        if node_id in manifest.exposures:
            found = manifest.exposures[node_id]
            node = CllNode.build_cll_node(manifest, "exposures", node_id)
            if hasattr(found.depends_on, "nodes"):
                parent_list = found.depends_on.nodes

        if hasattr(manifest, "semantic_models") and node_id in manifest.semantic_models:
            found = manifest.semantic_models[node_id]
            node = CllNode.build_cll_node(manifest, "semantic_models", node_id)
            if hasattr(found.depends_on, "nodes"):
                parent_list = found.depends_on.nodes

        if node_id in manifest.metrics:
            found = manifest.metrics[node_id]
            node = CllNode.build_cll_node(manifest, "metrics", node_id)
            if hasattr(found.depends_on, "nodes"):
                parent_list = found.depends_on.nodes

        return node, parent_list

    def get_manifests_by_id(self, unique_id: str):
        curr_manifest = self.get_manifest(base=False)
        base_manifest = self.get_manifest(base=True)
        if unique_id in curr_manifest.nodes.keys() or unique_id in base_manifest.nodes.keys():
            return {
                "current": curr_manifest.nodes.get(unique_id),
                "base": base_manifest.nodes.get(unique_id),
            }
        return None

    def build_name_to_unique_id_index(self) -> Dict[str, str]:
        name_to_unique_id = {}
        curr_manifest = self.get_manifest(base=False)
        base_manifest = self.get_manifest(base=True)

        for unique_id, node in base_manifest.nodes.items():
            name_to_unique_id[node.name] = unique_id
        for unique_id, node in curr_manifest.nodes.items():
            name_to_unique_id[node.name] = unique_id
        return name_to_unique_id

    def start_monitor_artifacts(self, callback: Callable = None):
        if self.artifacts_files:
            event_handler = ArtifactsEventHandler(self.artifacts_files, callback=callback)
            if self.target_path:
                self.artifacts_observer.schedule(event_handler, self.target_path, recursive=False)
            if self.base_path:
                self.artifacts_observer.schedule(event_handler, self.base_path, recursive=False)
            self.artifacts_observer.start()
            logger.info("Start monitoring dbt artifacts")

    def stop_monitor_artifacts(self):
        if self.artifacts_files:
            self.artifacts_observer.stop()
            self.artifacts_observer.join()
            logger.info("Stop monitoring artifacts")

    def start_monitor_base_env(self, callback: Callable = None):
        target_base_dir = os.path.join(self.runtime_config.project_root, "target-base")
        base_env_files = {
            os.path.join(target_base_dir, "manifest.json"),
            os.path.join(target_base_dir, "catalog.json"),
        }
        event_handler = EnvironmentEventHandler(self.base_env_observer, base_env_files, callback=callback)
        self.base_env_observer.schedule(event_handler, self.runtime_config.project_root, recursive=True)
        self.base_env_observer.start()
        logger.info("Start monitoring base environment")

    def stop_monitor_base_env(self):
        if self.base_env_observer.is_alive():
            self.base_env_observer.stop()
        self.base_env_observer.join()
        logger.info("Stop monitoring base environment")

    def set_artifacts(
        self,
        base_manifest: WritableManifest,
        curr_manifest: WritableManifest,
        manifest: Manifest,
        previous_manifest: Manifest,
        base_catalog: CatalogArtifact,
        curr_catalog: CatalogArtifact,
    ):
        self.curr_manifest = curr_manifest
        self.base_manifest = base_manifest
        self.manifest = manifest
        self.curr_catalog = curr_catalog
        self.base_catalog = base_catalog
        self.previous_state = previous_state(
            Path(self.base_path),
            Path(self.runtime_config.target_path),
            Path(self.runtime_config.project_root),
        )
        self.previous_state.manifest = previous_manifest

        # The dependencies of the review mode is derived from manifests.
        # It is a workaround solution to use macro dispatch
        # see: https://docs.getdbt.com/reference/dbt-jinja-functions/dispatch
        dependencies = {}
        for macro in self.manifest.macros.values():
            if macro.package_name not in dependencies:
                dependencies[macro.package_name] = self.runtime_config
        self.runtime_config.dependencies = dependencies

    def refresh(self, refresh_file_path: str = None):
        # Refresh the artifacts
        if refresh_file_path is None:
            return self.load_artifacts()

        # In single environment mode (target_path is equal to base_path),
        # we capture the original manifest as base and only update the current
        target_type = os.path.basename(os.path.dirname(refresh_file_path))
        if self.target_path and target_type == os.path.basename(self.target_path):
            if refresh_file_path.endswith("manifest.json"):
                self.curr_manifest = load_manifest(path=refresh_file_path)
                self.manifest = as_manifest(self.curr_manifest)
                self.get_cll_cached.cache_clear()
                self.get_change_analysis_cached.cache_clear()
            elif refresh_file_path.endswith("catalog.json"):
                self.curr_catalog = load_catalog(path=refresh_file_path)
                self.get_cll_cached.cache_clear()
                self.get_change_analysis_cached.cache_clear()
        elif self.base_path and target_type == os.path.basename(self.base_path):
            if refresh_file_path.endswith("manifest.json"):
                self.base_manifest = load_manifest(path=refresh_file_path)
                self.get_change_analysis_cached.cache_clear()
            elif refresh_file_path.endswith("catalog.json"):
                self.base_catalog = load_catalog(path=refresh_file_path)
                self.get_change_analysis_cached.cache_clear()

    def create_relation(self, model, base=False):
        node = self.find_node_by_name(model, base)
        if node is None:
            return None

        return self.adapter.Relation.create_from(self.runtime_config, node)

    def select_nodes(
        self,
        select: Optional[str] = None,
        exclude: Optional[str] = None,
        packages: Optional[list[str]] = None,
        view_mode: Optional[Literal["all", "changed_models"]] = None,
    ) -> Set[str]:
        import dbt.compilation
        from dbt.compilation import Compiler
        from dbt.graph import (
            NodeSelector,
            SelectionIntersection,
            SelectionUnion,
            parse_difference,
        )

        select_list = [select] if select else None
        exclude_list = [exclude] if exclude else None

        def _parse_difference(include, exclude):
            if dbt_version < "v1.8":
                return parse_difference(include, exclude, "eager")
            else:
                return parse_difference(include, exclude)

        specs = [_parse_difference(select_list, exclude_list)]

        if packages is not None:
            package_spec = SelectionUnion([_parse_difference([f"package:{p}"], None) for p in packages])
            specs.append(package_spec)
        if view_mode and view_mode == "changed_models":
            specs.append(_parse_difference(["1+state:modified+"], None))
        spec = SelectionIntersection(specs)

        manifest = Manifest()
        manifest.metadata.adapter_type = self.adapter.type()
        manifest_prev = self.previous_state.manifest
        manifest_curr = self.manifest

        manifest.nodes = {**manifest_curr.nodes}
        # # mark a node is removed if the node id is no in the curr nodes
        for node_id, node in manifest_prev.nodes.items():
            if node_id not in manifest.nodes:
                node_dict = node.to_dict()
                if "raw_code" in node_dict:
                    node_dict["raw_code"] = "__removed__"
                node_class = type(node)
                removed_node = node_class.from_dict(node_dict)
                manifest.nodes[node_id] = removed_node

        manifest.macros = {**manifest_prev.macros, **manifest_curr.macros}
        manifest.sources = {**manifest_prev.sources, **manifest_curr.sources}
        manifest.exposures = {**manifest_prev.exposures, **manifest_curr.exposures}
        manifest.metrics = {**manifest_prev.metrics, **manifest_curr.metrics}
        if hasattr(manifest_prev, "semantic_models"):
            manifest.semantic_models = {
                **manifest_prev.semantic_models,
                **manifest_curr.semantic_models,
            }

        compiler = Compiler(self.runtime_config)
        # disable to print compile states
        tmp_func = dbt.compilation.print_compile_stats
        dbt.compilation.print_compile_stats = lambda x: None
        graph = compiler.compile(manifest, write=False)
        dbt.compilation.print_compile_stats = tmp_func
        selector = NodeSelector(graph, manifest, previous_state=self.previous_state)

        # disable "The selection criterion does not match"
        with silence_no_nodes_warning():
            return selector.get_selected(spec)

    def export_artifacts(self) -> ArtifactsRoot:
        """
        Export the artifacts from the current state
        """
        artifacts = ArtifactsRoot()

        def _load_artifact(artifact):
            return artifact.to_dict() if artifact else None

        artifacts.base = {
            "manifest": _load_artifact(self.base_manifest),
            "catalog": _load_artifact(self.base_catalog),
        }
        artifacts.current = {
            "manifest": _load_artifact(self.curr_manifest),
            "catalog": _load_artifact(self.curr_catalog),
        }
        return artifacts

    def export_artifacts_from_file(self) -> ArtifactsRoot:
        """
        Export the artifacts from the state file. This is the old implementation
        """
        artifacts = ArtifactsRoot()
        target_path = self.runtime_config.target_path
        target_base_path = self.base_path

        def _load_artifact(path):
            if not os.path.isfile(path):
                return None

            with open(path, "r", encoding="utf-8") as f:
                json_content = f.read()
                return json.loads(json_content)

        project_root = self.runtime_config.project_root
        artifacts.base = {
            "manifest": _load_artifact(os.path.join(project_root, target_base_path, "manifest.json")),
            "catalog": _load_artifact(os.path.join(project_root, target_base_path, "catalog.json")),
        }
        artifacts.current = {
            "manifest": _load_artifact(os.path.join(project_root, target_path, "manifest.json")),
            "catalog": _load_artifact(os.path.join(project_root, target_path, "catalog.json")),
        }
        return artifacts

    def import_artifacts(self, artifacts: ArtifactsRoot, merge=True):
        # Merge the artifacts from the state file or cloud
        def _select_artifact(
            original: Union[WritableManifest, CatalogArtifact],
            new: Union[WritableManifest, CatalogArtifact],
        ):
            if merge:
                if not original:
                    return new
                if not new:
                    return original
                return original if original.metadata.generated_at > new.metadata.generated_at else new
            else:
                return new

        self.base_manifest = _select_artifact(self.base_manifest, load_manifest(data=artifacts.base.get("manifest")))
        self.curr_manifest = _select_artifact(self.curr_manifest, load_manifest(data=artifacts.current.get("manifest")))
        self.base_catalog = _select_artifact(self.base_catalog, load_catalog(data=artifacts.base.get("catalog")))
        self.curr_catalog = _select_artifact(self.curr_catalog, load_catalog(data=artifacts.current.get("catalog")))

        self.manifest = as_manifest(self.curr_manifest)
        self.previous_state = previous_state(
            Path(self.base_path),
            Path(self.runtime_config.target_path),
            Path(self.runtime_config.project_root),
        )
        self.previous_state.manifest = as_manifest(self.base_manifest)

        # The dependencies of the review mode is derived from manifests.
        # It is a workaround solution to use macro dispatch
        # see: https://docs.getdbt.com/reference/dbt-jinja-functions/dispatch
        dependencies = {}
        for macro in self.manifest.macros.values():
            if macro.package_name not in dependencies:
                dependencies[macro.package_name] = self.runtime_config

        self.runtime_config.dependencies = dependencies

        if not self.curr_manifest or not self.base_manifest:
            raise Exception(
                "No enough dbt artifacts in the state file. Please use the latest recce to generate the recce state"
            )

    @contextmanager
    def connection_named(self, name: str) -> Iterator[None]:
        with self.adapter.connection_named(name):
            yield

    def get_thread_connection(self) -> Connection:
        return self.adapter.connections.get_thread_connection()

    def cancel(self, connection: Connection):
        self.adapter.connections.cancel(connection)

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()



import pytest

from recce.tasks.histogram import (
    HistogramDiffCheckValidator,
    HistogramDiffTask,
    _is_histogram_supported,
)


def test_histogram(dbt_test_helper):
    csv_data = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        4,Dolly,50
        """

    dbt_test_helper.create_model("customers", csv_data, csv_data)

    params = {"model": "customers", "column_name": "age", "column_type": "int"}

    task = HistogramDiffTask(params)
    run_result = task.execute()

    # {
    #     'base': {'counts': [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], 'total': 4},
    #     'current': {'counts': [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1], 'total': 4},
    #     'min': 25, 'max': 50,
    #     'bin_edges': [25, 26, ..., 51],
    #     'labels': ['25-26', ..., '51-52']
    # }
    assert run_result["current"]["counts"][0] == 1
    assert run_result["current"]["counts"][-1] == 1
    assert run_result["current"]["total"] == 4
    assert run_result["min"] == 25
    assert run_result["max"] == 50
    assert run_result["bin_edges"][0] == 25
    assert run_result["bin_edges"][-1] == 51


def test_histogram_emtpy(dbt_test_helper):
    csv_data = """
    customer_id,name,age
    1,Alice,30
    2,Bob,25
    3,Charlie,35
    4,Dolly,50
    """

    csv_data_zero = """
    customer_id,name,age
    """

    dbt_test_helper.create_model("customers", csv_data_zero, csv_data_zero)
    dbt_test_helper.create_model("customers2", csv_data, csv_data_zero)
    dbt_test_helper.create_model("customers3", csv_data_zero, csv_data)

    params = {"model": "customers", "column_name": "age", "column_type": "int"}

    task = HistogramDiffTask(params)
    run_result = task.execute()

    assert len(run_result["current"]["counts"]) == 0
    assert run_result["current"]["total"] == 0
    assert run_result["min"] is None
    assert run_result["max"] is None
    assert len(run_result["bin_edges"]) == 0

    params = {"model": "customers2", "column_name": "age", "column_type": "int"}

    task = HistogramDiffTask(params)
    run_result = task.execute()
    assert run_result["base"]["counts"][0] == 1
    assert run_result["base"]["counts"][-1] == 1
    assert run_result["base"]["total"] == 4
    assert run_result["current"]["counts"][0] == 0
    assert run_result["current"]["counts"][-1] == 0
    assert run_result["current"]["total"] == 0
    assert run_result["min"] == 25
    assert run_result["max"] == 50
    assert run_result["bin_edges"][0] == 25
    assert run_result["bin_edges"][-1] == 51

    params = {"model": "customers3", "column_name": "age", "column_type": "int"}

    task = HistogramDiffTask(params)
    run_result = task.execute()
    assert run_result["base"]["counts"][0] == 0
    assert run_result["base"]["counts"][-1] == 0
    assert run_result["base"]["total"] == 0
    assert run_result["current"]["counts"][0] == 1
    assert run_result["current"]["counts"][-1] == 1
    assert run_result["current"]["total"] == 4
    assert run_result["min"] == 25
    assert run_result["max"] == 50
    assert run_result["bin_edges"][0] == 25
    assert run_result["bin_edges"][-1] == 51


def test_validator():
    def validate(params: dict = {}, view_options: dict = {}):
        HistogramDiffCheckValidator().validate(
            {
                "name": "test",
                "type": "histogram_diff",
                "params": params,
                "view_options": view_options,
            }
        )

    validate(
        {
            "model": "customers",
            "column_name": "age",
            "column_type": "int",
        }
    )

    with pytest.raises(ValueError):
        validate({})


def test_is_column_type_supported_by_histogram():
    assert _is_histogram_supported("varchar") is False
    assert _is_histogram_supported("varchar(16)") is False
    assert _is_histogram_supported("varchar(256)") is False
    assert _is_histogram_supported("bool") is False
    assert _is_histogram_supported("int") is True



import pytest


def test_default_validator():
    from recce.tasks.core import CheckValidator

    CheckValidator().validate(
        {
            "name": "test",
            "type": "row_count_diff",
            "params": {},
        }
    )

    with pytest.raises(ValueError):
        CheckValidator().validate(
            {
                "type": "row_count_diff",
            }
        )


def test_query_diff_validator():
    from recce.tasks.query import QueryDiffCheckValidator

    QueryDiffCheckValidator().validate(
        {
            "name": "test",
            "type": "query_diff",
            "params": {"sql_template": "select * from {{ model }}"},
        }
    )
    with pytest.raises(ValueError):
        QueryDiffCheckValidator().validate(
            {
                "name": "test",
                "type": "query_diff",
            }
        )


def test_lineage_diff_validator():
    from recce.tasks.lineage import LineageDiffCheckValidator

    LineageDiffCheckValidator().validate(
        {
            "name": "test",
            "type": "lineage_diff",
            "params": {},
        }
    )
    LineageDiffCheckValidator().validate(
        {
            "name": "test",
            "type": "lineage_diff",
            "view_options": {},
        }
    )
    with pytest.raises(ValueError):
        LineageDiffCheckValidator().validate(
            {
                "type": "lineage_diff",
            }
        )



from enum import Enum
from typing import Union

from recce.util.recce_cloud import (
    get_recce_cloud_onboarding_state,
    set_recce_cloud_onboarding_state,
)


class OnboardingState(Enum):
    SIGNUP_SUCCESSFUL = "signup_successful"
    LAUNCHED_WITH_TOKEN = "launched_with_token"
    CONFIGURE_TWO_ENV = "configure_two_env"


def update_onboarding_state(api_token: Union[str, bool, None], is_single_env: bool) -> OnboardingState:
    if api_token:
        # existing onboarding_state values -> new, launched, launched_with_two_envs
        #   new -> launched
        #   new -> launched_with_two_envs
        #   launched -> launched_with_two_envs
        cloud_onboarding_state = get_recce_cloud_onboarding_state(api_token)

        if cloud_onboarding_state == OnboardingState.SIGNUP_SUCCESSFUL.value:
            # User has an API Token and is a "new" user
            if is_single_env:
                # Mark the onboarding state as "launched" if the user is new
                set_recce_cloud_onboarding_state(api_token, OnboardingState.LAUNCHED_WITH_TOKEN.value)
                return OnboardingState.LAUNCHED_WITH_TOKEN
            else:
                set_recce_cloud_onboarding_state(api_token, OnboardingState.CONFIGURE_TWO_ENV.value)
                return OnboardingState.CONFIGURE_TWO_ENV
        elif cloud_onboarding_state == OnboardingState.LAUNCHED_WITH_TOKEN.value:
            # User has an API Token and has Two Environments
            if is_single_env:
                # Just return the current state
                return OnboardingState.LAUNCHED_WITH_TOKEN
            else:
                set_recce_cloud_onboarding_state(api_token, OnboardingState.CONFIGURE_TWO_ENV.value)
                return OnboardingState.CONFIGURE_TWO_ENV
        elif cloud_onboarding_state == OnboardingState.CONFIGURE_TWO_ENV.value:
            # Just return the current state
            return OnboardingState.CONFIGURE_TWO_ENV

    return OnboardingState.SIGNUP_SUCCESSFUL



import pytest


def test_validator():
    from recce.tasks.lineage import LineageDiffCheckValidator

    validator = LineageDiffCheckValidator()

    def validate(params: dict):
        validator.validate(
            {
                "name": "test",
                "type": "schema_diff",
                "params": params,
            }
        )

    # Select all models
    validate({})

    # Select by selector
    validate(
        {
            "select": "customers",
            "exclude": "customers",
            "packages": ["jaffle_shop"],
            "view_mode": "all",
        }
    )

    # packages should be an array
    with pytest.raises(ValueError):
        validate(
            {
                "packages": "jaffle_shop",
            }
        )

    # view_mode should be 'all' or 'changed_models'
    validate(
        {
            "view_mode": None,
        }
    )
    validate(
        {
            "view_mode": "all",
        }
    )
    with pytest.raises(ValueError):
        validate(
            {
                "view_mode": "abc",
            }
        )



#!/usr/bin/env python
import os
from distutils.core import setup

from setuptools import find_packages  # type: ignore


def _get_version():
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), "recce", "VERSION"))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


setup(
    name="recce",
    version=_get_version(),
    description="Environment diff tool for dbt",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="InfuseAI Dev Team",
    author_email="dev@infuseai.io",
    url="https://github.com/InfuseAI/recce",
    entry_points={"console_scripts": ["recce = recce.cli:cli"]},
    python_requires=">=3.9",
    packages=find_packages(),
    install_requires=[
        "boto3",
        "requests>=2.28.1",
        "ruamel.yaml>=0.18.6",
        "click>=7.1",
        "deepdiff>=7.0,<9.0",
        "portalocker",
        "fastapi",
        "itsdangerous",
        "uvicorn",
        "pydantic",
        "jinja2",
        "requests>=2.28.1",
        "rich>=12.0.0",
        "sentry-sdk",
        "watchdog",
        "websockets",
        "py-markdown-table",
        "python-dateutil",
        "python-multipart",
        "GitPython",
        "PyGithub",
        "sqlglot",
        "pytz",
        "packaging",
    ],
    tests_require=["pytest"],
    extras_require={
        "mcp": [
            "mcp>=1.0.0",
        ],
        "dev": [
            "pytest>=4.6",
            "pytest-asyncio",
            "pytest-flake8",
            "black>=25.1.0",
            "isort>=6.0.1",
            "flake8>=7.2.0",
            "pre-commit>=4.2.0",
            "pytest-mypy",
            "pytest-cov",
            "twine",
            "tox",
            "pandas",
            "httpx",
        ],
    },
    project_urls={
        "Bug Tracker": "https://github.com/InfuseAI/recce/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    package_data={"recce": ["VERSION", "data/**", "event/SENTRY_DNS", "event/CONFIG"]},
)



import asyncio
import json
import logging
import os
import signal
import uuid
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, Set

from fastapi import (
    BackgroundTasks,
    FastAPI,
    Form,
    HTTPException,
    Request,
    Response,
    UploadFile,
    WebSocket,
)
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse
from fastapi.staticfiles import StaticFiles
from pydantic import BaseModel, ValidationError
from pytz import utc
from starlette.middleware.gzip import GZipMiddleware
from starlette.middleware.sessions import SessionMiddleware
from starlette.websockets import WebSocketDisconnect

from . import __latest_version__, __version__, event, is_recce_cloud_instance
from .apis.check_api import check_router
from .apis.run_api import run_router
from .config import RecceConfig
from .connect_to_cloud import (
    connect_to_cloud_background_task,
    generate_key_pair,
    get_connection_url,
    is_callback_server_running,
    prepare_connection_url,
)
from .core import RecceContext, default_context, load_context
from .event import get_recce_api_token, log_api_event, log_single_env_event
from .exceptions import RecceException
from .github import is_github_codespace
from .models.types import CllData
from .run import load_preset_checks
from .state import RecceShareStateManager, RecceStateLoader

logger = logging.getLogger("uvicorn")


class RecceServerMode(str, Enum):
    server = "server"
    preview = "preview"
    read_only = "read-only"

    def __str__(self):
        return self.value

    @staticmethod
    def available_members() -> Set[str]:
        return ["server", "preview", "read-only"]


@dataclass
class AppState:
    command: Optional[str] = None
    state_loader: Optional[RecceStateLoader] = None
    kwargs: Optional[dict] = None
    flag: Optional[dict] = None
    auth_options: Optional[dict] = None
    lifetime: Optional[int] = None
    lifetime_expired_at: Optional[datetime] = None
    share_url: Optional[str] = None
    host: Optional[str] = None
    port: Optional[int] = None


def schedule_lifetime_termination(app_state):
    def terminating_server():
        pid = os.getpid()
        logger.info(f"Terminating server process [{pid}] manually")
        os.kill(pid, signal.SIGINT)

    # Terminate the server process after the specified lifetime
    logger.info(f"[Configuration] The lifetime of the server is {app_state.lifetime} seconds")
    app.state.lifetime_expired_at = datetime.now(utc) + timedelta(seconds=app_state.lifetime)
    asyncio.get_running_loop().call_later(app_state.lifetime, terminating_server)


def setup_server(app_state: AppState) -> RecceContext:
    from rich.console import Console

    from .core import load_context

    console = Console()
    state_loader = app_state.state_loader
    kwargs = app_state.kwargs
    ctx = load_context(**kwargs, state_loader=state_loader)
    ctx.start_monitor_artifacts(callback=dbt_artifacts_updated_callback)
    single_env = False
    if app_state.flag.get("single_env_onboarding", False) is True:
        # [Experiment 2] Start with Single Environment
        single_env = True
        ctx.start_monitor_base_env(callback=dbt_env_updated_callback)
        log_single_env_event()

    # Initialize Recce Config
    config = RecceConfig(config_file=kwargs.get("config"))
    if state_loader.state is None:
        preset_checks = config.get("checks", [])
        if preset_checks and len(preset_checks) > 0:
            console.rule("Loading Preset Checks")
            load_preset_checks(preset_checks)

    from recce.event import log_load_state

    log_load_state(command="server", single_env=single_env)

    return ctx


def teardown_server(app_state: AppState, ctx: RecceContext):
    # pull latest state, merge runs/checks and pick the newer artifacts
    state_loader = ctx.state_loader
    state_loader.refresh()
    if state_loader.state:
        ctx.import_state(state_loader.state, merge=True)
    state_loader.export(ctx.export_state())
    ctx.stop_monitor_artifacts()
    if app_state.flag.get("single_env_onboarding", False):
        ctx.stop_monitor_base_env()


def setup_ready_only(app_state: AppState):
    pass


def teardown_ready_only(app_state: AppState):
    pass


def setup_preview(app_state: AppState):
    state_loader = app_state.state_loader
    kwargs = app_state.kwargs
    ctx = load_context(**kwargs, state_loader=state_loader)
    return ctx


def teardown_preview(app_state: AppState, ctx: RecceContext):
    state_loader = app_state.state_loader
    state_loader.export(ctx.export_state())
    pass


@asynccontextmanager
async def lifespan(fastapi: FastAPI):
    ctx = None
    app_state: AppState = app.state

    if app_state.command == "server":
        ctx = setup_server(app_state)
    elif app_state.command == "read-only":
        setup_ready_only(app_state)
    elif app_state.command == "preview":
        ctx = setup_preview(app_state)

    if app_state.lifetime is not None and app_state.lifetime > 0:
        schedule_lifetime_termination(app_state)

    yield

    if app_state.command == "server":
        teardown_server(app_state, ctx)
    elif app_state.command == "read_only":
        teardown_ready_only(app_state)
    elif app_state.command == "preview":
        teardown_preview(app_state, ctx)


app = FastAPI(lifespan=lifespan)


def verify_json_file(file_path: str) -> bool:
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            json.load(f)
    except Exception:
        return False
    return True


def dbt_artifacts_updated_callback(file_changed_event: Any):
    src_path = Path(file_changed_event.src_path)
    target_type = src_path.parent.name
    file_name = src_path.name

    if not verify_json_file(file_changed_event.src_path):
        logger.debug("Skip to refresh the artifacts because the file is not updated completely.")
        return

    logger.info(f"Detect {target_type} file {file_changed_event.event_type}: {file_name}")
    ctx = load_context()
    ctx.refresh_manifest(file_changed_event.src_path)
    broadcast_command = {
        "command": "refresh",
        "event": {"eventType": file_changed_event.event_type, "srcPath": file_changed_event.src_path},
    }
    payload = json.dumps(broadcast_command)
    asyncio.run(broadcast(payload))


def dbt_env_updated_callback():
    logger.info("Detect 'manifest.json' and 'catalog.json' are generated under 'target-base' directory")
    broadcast_command = {
        "command": "relaunch",
    }
    payload = json.dumps(broadcast_command)
    asyncio.run(broadcast(payload))


clients = set()

origins = [
    "http://localhost:3000",
    "http://localhost:3001",
]

app.add_middleware(GZipMiddleware, minimum_size=1000)
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(
    SessionMiddleware,
    secret_key=uuid.uuid4(),
)


@app.middleware("http")
async def set_context_by_cookie(request: Request, call_next):
    response = await call_next(request)

    user_id_in_cookie = request.cookies.get("recce_user_id")
    user_id = event.get_user_id()

    if event.is_anonymous_tracking() is False:
        # Disable anonymous tracking
        user_id = None

    if user_id_in_cookie is None or user_id_in_cookie != user_id:
        response.set_cookie(key="recce_user_id", value=user_id)
    return response


@app.middleware("http")
async def disable_cache(request: Request, call_next):
    response = await call_next(request)

    # disable cache for '/' and '/index.html'
    if request.url.path in ["/", "/index.html"]:
        response.headers["Cache-Control"] = "no-store"

    return response


@app.get("/api/health")
async def health_check(request: Request):
    return {"status": "ok"}


class RecceInstanceInfoOut(BaseModel):
    server_mode: RecceServerMode
    read_only: bool
    preview: bool
    single_env: bool
    authed: bool
    cloud_instance: bool
    lifetime_expired_at: Optional[datetime] = None
    share_url: Optional[str] = None
    session_id: Optional[str] = None


@app.get("/api/instance-info", response_model=RecceInstanceInfoOut, response_model_exclude_none=True)
async def recce_instance_info():
    app_state: AppState = app.state
    flag = app_state.flag
    read_only = flag.get("read_only", False)
    single_env = flag.get("single_env_onboarding", False)

    api_token = get_recce_api_token()

    return {
        "server_mode": app_state.command,
        "read_only": read_only,
        "preview": flag.get("preview", False),
        "single_env": single_env,
        "authed": True if api_token else False,
        "cloud_instance": is_recce_cloud_instance(),
        "lifetime_expired_at": app_state.lifetime_expired_at,  # UTC timezone
        "share_url": app_state.share_url,
        "session_id": app_state.state_loader.session_id if app_state.state_loader else None,
        # TODO: Add more instance info which won't change during the instance lifecycle
        # review_mode
        # cloud_mode
        # demo
        # single env
    }


@app.get("/api/flag")
async def config_flag():
    app_state: AppState = app.state
    flag = app_state.flag
    return flag


@app.post("/api/relaunch-hint/completed", status_code=204)
async def mark_relaunch_hint_completed():
    app.state.flag["show_relaunch_hint"] = False


@app.get("/api/info")
async def get_info():
    """
    Get the information of the current context.
    """
    context = default_context()
    demo = os.environ.get("DEMO", False)
    is_codespace = is_github_codespace()

    if demo:
        state = context.export_demo_state()
    else:
        state = context.export_state()

    support_tasks = context.support_tasks()
    if context.state_loader and context.state_loader.state_file:
        filename = os.path.basename(context.state_loader.state_file)
    else:
        filename = None

    state_metadata = context.state_loader.state.metadata if context.state_loader.state else None
    lineage_diff = context.get_lineage_diff()

    try:
        info = {
            "state_metadata": state_metadata,
            "adapter_type": context.adapter_type,
            "review_mode": context.review_mode,
            "git": state.git.to_dict() if state.git else None,
            "pull_request": state.pull_request.to_dict() if state.pull_request else None,
            "lineage": lineage_diff,
            "demo": bool(demo),
            "codespace": bool(is_codespace),
            "cloud_mode": context.state_loader.cloud_mode,
            "file_mode": context.state_loader.state_file is not None,
            "filename": filename,
            "support_tasks": support_tasks,
        }

        if context.adapter_type == "sqlmesh":
            from recce.adapter.sqlmesh_adapter import SqlmeshAdapter

            sqlmesh_adapter: SqlmeshAdapter = context.adapter
            info["sqlmesh"] = {
                "base_env": sqlmesh_adapter.base_env.name,
                "current_env": sqlmesh_adapter.curr_env.name,
            }

        return info
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


class CllIn(BaseModel):
    node_id: Optional[str] = None
    column: Optional[str] = None
    change_analysis: Optional[bool] = False
    no_cll: Optional[bool] = False
    no_upstream: Optional[bool] = False
    no_downstream: Optional[bool] = False


class CllOutput(BaseModel):
    current: CllData


@app.post("/api/cll", response_model=CllOutput)
async def column_level_lineage_by_node(cll_input: CllIn):
    from recce.adapter.dbt_adapter import DbtAdapter

    dbt_adapter: DbtAdapter = default_context().adapter
    cll = dbt_adapter.get_cll(
        node_id=cll_input.node_id,
        column=cll_input.column,
        change_analysis=cll_input.change_analysis,
        no_upstream=cll_input.no_upstream,
        no_downstream=cll_input.no_downstream,
        no_cll=cll_input.no_cll,
    )

    return CllOutput(current=cll)


class SelectNodesInput(BaseModel):
    select: Optional[str] = None
    exclude: Optional[str] = None
    packages: Optional[list[str]] = None
    view_mode: Optional[Literal["all", "changed_models"]] = None


class SelectNodesOutput(BaseModel):
    nodes: Set[str] = []


@app.post("/api/select", response_model=SelectNodesOutput)
async def select_nodes(input: SelectNodesInput):
    context = default_context()

    if context.adapter_type != "dbt":
        raise HTTPException(status_code=400, detail="Only dbt adapter is supported")

    try:
        nodes = context.adapter.select_nodes(
            select=input.select,
            exclude=input.exclude,
            packages=input.packages,
            view_mode=input.view_mode,
        )
        nodes = [node for node in nodes if not node.startswith("test.")]
        return SelectNodesOutput(nodes=nodes)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/api/model/{model_id}")
async def get_columns(model_id: str):
    context = default_context()
    try:
        return {
            "model": {
                "base": context.get_model(model_id, base=True),
                "current": context.get_model(model_id, base=False),
            }
        }
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.post("/api/save", response_class=PlainTextResponse, status_code=200)
async def save_handler():
    """
    Save the in-memory state to the state file
    """
    try:
        # Sync the state file
        context = default_context()
        log_api_event("save", dict(state_loader_mode=context.state_loader_mode()))
        state_loader = context.state_loader
        if not state_loader.cloud_mode and state_loader.state_file is None:
            raise RecceException("Not file mode or cloud mode")

        context.sync_state("overwrite")
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


class SaveAsOrRenameInput(BaseModel):
    # The filename. The filename should not contain directory.
    filename: Optional[str] = None
    # Overwrite the file if it exists
    overwrite: Optional[bool] = False


def saveas_or_rename(input: SaveAsOrRenameInput, rename: bool = False):
    context = default_context()
    state_loader = context.state_loader
    if state_loader.cloud_mode:
        raise RecceException("Cloud mode does not support rename")

    new_filename = input.filename
    if os.path.dirname(new_filename):
        raise RecceException("The new filename should not contain directory")
    if not new_filename.endswith(".json"):
        raise RecceException("The new filename should end with .json")

    old_path = state_loader.state_file
    if old_path:
        old_dir = os.path.dirname(state_loader.state_file)
        old_filename = os.path.basename(state_loader.state_file)
        if old_filename == new_filename:
            raise RecceException("The new filename is the same as the current filename")
        new_path = os.path.join(old_dir, new_filename)
    else:
        new_path = new_filename

    if os.path.exists(new_path):
        if os.path.isdir(new_path):
            raise HTTPException(status_code=400, detail=f"The file {new_path} exists and is a directory")

        if not input.overwrite:
            raise HTTPException(status_code=409, detail=f"The file {new_filename} already exists")

    state_loader.state_file = new_path
    context.sync_state("overwrite")
    if rename and os.path.exists(old_path):
        os.remove(old_path)


@app.post("/api/save-as", response_class=PlainTextResponse, status_code=200)
async def save_as_handler(input: SaveAsOrRenameInput):
    """
    Save the state to a new file
    """
    context = default_context()
    try:
        log_api_event("saveas", dict(state_loader_mode=context.state_loader_mode()))
        saveas_or_rename(input, rename=False)
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


@app.post("/api/rename", response_class=PlainTextResponse, status_code=200)
async def rename_handler(input: SaveAsOrRenameInput):
    """
    Rename the state to a new file
    """
    context = default_context()
    try:
        log_api_event("rename", dict(state_loader_mode=context.state_loader_mode()))
        saveas_or_rename(input, rename=True)
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


@app.post("/api/export", response_class=PlainTextResponse, status_code=200)
async def export_handler():
    """
    Export the recce state to the client.
    """
    context = default_context()
    try:
        log_api_event("export", dict(state_loader_mode=context.state_loader_mode()))
        return context.export_state().to_json()
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


@app.post("/api/import", status_code=200)
async def import_handler(
    file: Annotated[UploadFile, Form()], checks_only: Annotated[bool, Form()], background_tasks: BackgroundTasks
):
    """
    Import the recce state from the client.
    """
    from recce.state import RecceState

    context = default_context()
    try:
        log_api_event("import", dict(state_loader_mode=context.state_loader_mode()))
        content = await file.read()
        state = RecceState.from_json(content)

        if checks_only:
            import_checks = context.import_checks(state)
            background_tasks.add_task(context.sync_state, "overwrite")
            return {"runs": 0, "checks": import_checks}

        import_runs, import_checks = context.import_state(state)
        return {"runs": import_runs, "checks": import_checks}
    except ValidationError as e:
        raise HTTPException(status_code=400, detail=str(e))
    except RecceException as e:
        raise HTTPException(status_code=400, detail=e.message)


class SyncStateInput(BaseModel):
    method: Optional[str] = None


@app.post("/api/sync", status_code=202)
async def sync_handler(input: SyncStateInput, response: Response, background_tasks: BackgroundTasks):
    """
    Sync the state with the external storage. (two-way sync)

    This is used to sync the state with the external (local or cloud) storage. There are three methods:
    - overwrite: Overwrite the external storage with the in-memory state.
    - revert: Revert the in-memory state with the external storage.
    - merge: Merge the state between the in-memory and external storage.
    """
    context = default_context()
    state_loader = context.state_loader
    method = input.method
    log_api_event(
        "sync",
        dict(
            state_loader_mode=context.state_loader_mode(),
            method=method,
        ),
    )

    if not method:
        is_conflict = state_loader.check_conflict()
        if is_conflict:
            raise HTTPException(status_code=409, detail="Conflict detected")
        method = "overwrite"

    is_syncing = state_loader.state_lock.locked()
    if is_syncing:
        response.status_code = 208
        return {"status": "syncing"}

    def reload_state():
        ctx = default_context()
        ctx.sync_state(method)

    background_tasks.add_task(reload_state)
    response.status_code = 202
    return {"status": "request accepted"}


@app.get("/api/sync", status_code=200)
async def sync_status(response: Response):
    """
    Get the sync status.
    """
    context = default_context()
    if context.state_loader.state_lock.locked():
        response.status_code = 208
        return {"status": "syncing"}

    response.status_code = 200
    return {"status": "idle"}


class ShareStateOutput(BaseModel):
    status: str
    message: str
    share_url: Optional[str] = None


@app.post("/api/share", response_model=ShareStateOutput)
async def share_state():
    """
    Share the recce state to the external storage. (one-way sync)
    """
    app_state: AppState = app.state
    state_manager = RecceShareStateManager(app_state.auth_options)
    if not state_manager.verify():
        error, hint = state_manager.error_and_hint
        raise HTTPException(status_code=400, detail=f"Failed to share state: {error}. {hint}")

    context = default_context()
    state_loader = context.state_loader

    file_name = "recce_state.json"
    if state_loader.state_file:
        file_name = os.path.basename(state_loader.state_file)

    state = state_loader.state
    if state_loader.state is None:
        state = context.export_state()

    response = state_manager.share_state(file_name, state)

    return ShareStateOutput(**response)


class VersionOut(BaseModel):
    version: str
    latestVersion: str


@app.get("/api/version", response_model=VersionOut)
async def version():
    try:
        return dict(
            version=__version__,
            latestVersion=__latest_version__,
        )
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.websocket("/api/ws")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    clients.add(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            if data == "ping":
                await websocket.send_text("pong")
    except WebSocketDisconnect:
        clients.remove(websocket)


async def broadcast(data: str):
    for client in clients:
        await client.send_text(data)


@app.post("/api/connect")
async def generate_connect_to_cloud_url(background_tasks: BackgroundTasks):
    if is_callback_server_running():
        return {"connection_url": get_connection_url()}

    private_key, public_key = generate_key_pair()
    connection_url, callback_port = prepare_connection_url(public_key)

    background_tasks.add_task(connect_to_cloud_background_task, private_key, callback_port, connection_url)
    return {
        "connection_url": connection_url,
    }


@app.get("/api/users")
async def get_user_info():
    from recce.connect_to_cloud import RecceCloud

    context = default_context()
    user_token = get_recce_api_token() or context.state_loader.token
    cloud = RecceCloud(user_token)
    try:
        user_info = cloud.get_user_info()
        return user_info
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


api_prefix = "/api"
app.include_router(check_router, prefix=api_prefix)
app.include_router(run_router, prefix=api_prefix)

static_folder_path = Path(__file__).parent / "data"
app.mount("/", StaticFiles(directory=static_folder_path, html=True), name="static")



from unittest.mock import patch

import pytest

from recce.core import set_default_context

from .dbt_test_helper import DbtTestHelper


@pytest.fixture
def dbt_test_helper():
    with patch("recce.adapter.dbt_adapter.log_performance"):
        helper = DbtTestHelper()
        context = helper.context
        set_default_context(context)
        yield helper
        helper.cleanup()



import base64
import unittest
from unittest.mock import patch
from urllib.parse import quote

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding

from recce.connect_to_cloud import (
    connect_to_cloud_background_task,
    decrypt_code,
    generate_key_pair,
    is_callback_server_running,
    prepare_connection_url,
)


class ConnectToCloudTests(unittest.TestCase):

    def test_generate_key_pair(self):
        private_key, public_key = generate_key_pair()
        self.assertIsNotNone(private_key)
        self.assertIsNotNone(public_key)
        self.assertEqual(private_key.public_key().public_numbers(), public_key.public_numbers())

    def test_prepare_connection_url(self):
        _, public_key = generate_key_pair()
        url, port = prepare_connection_url(public_key)
        self.assertIn("connect?", url)
        self.assertTrue(port >= 10000 and port <= 15000)

    def test_decrypt_code(self):
        private_key, public_key = generate_key_pair()
        test_string = "recce-api-token-123"
        ciphertext = public_key.encrypt(
            test_string.encode(),
            padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA1()), algorithm=hashes.SHA1(), label=None),
        )
        b64_ciphertext = base64.b64encode(ciphertext).decode()
        result = decrypt_code(private_key, b64_ciphertext)
        self.assertEqual(result, test_string)

    @patch("recce.connect_to_cloud.update_recce_api_token")
    @patch("recce.connect_to_cloud.update_onboarding_state")
    @patch("recce.connect_to_cloud.RecceCloud")
    def test_handle_callback_request_success(self, mock_recce_cloud, mock_update_state, mock_update_token):
        private_key, public_key = generate_key_pair()

        # Prepare encrypted token
        test_token = "recce-api-token-xyz"
        ciphertext = public_key.encrypt(
            test_token.encode(),
            padding.OAEP(mgf=padding.MGF1(algorithm=hashes.SHA1()), algorithm=hashes.SHA1(), label=None),
        )
        encrypted_b64 = base64.b64encode(ciphertext).decode()

        # Set up mocks
        mock_recce_cloud.return_value.verify_token.return_value = True

        from recce.connect_to_cloud import handle_callback_request

        result = handle_callback_request(f"code={quote(encrypted_b64)}", private_key)

        assert result == test_token
        mock_update_token.assert_called_once_with(test_token)
        mock_update_state.assert_called_once_with(test_token, False)

    def test_is_callback_server_running(self):
        # Should return False by default
        self.assertFalse(is_callback_server_running())

    @patch("recce.connect_to_cloud.run_one_time_http_server")
    def test_connect_to_cloud_background_task_runs(self, mock_server):
        private_key, public_key = generate_key_pair()
        url, port = prepare_connection_url(public_key)

        connect_to_cloud_background_task(private_key, port, url)
        mock_server.assert_called_once()


if __name__ == "__main__":
    unittest.main()



import json
import os
from typing import Optional, Union

import requests
from pydantic import BaseModel

from recce.git import hosting_repo
from recce.github import recce_pr_information
from recce.util.pydantic_model import pydantic_model_dump


class PullRequestInfo(BaseModel):
    id: Optional[Union[int, str]] = None
    title: Optional[str] = None
    url: Optional[str] = None
    branch: Optional[str] = None
    base_branch: Optional[str] = None
    repository: Optional[str] = None

    def to_dict(self):
        return pydantic_model_dump(self)


def fetch_pr_metadata(**kwargs):
    pr_info = PullRequestInfo()

    # fetch from github action event path
    metadata = fetch_pr_metadata_from_event_path()
    if metadata is not None:
        pr_info.id = metadata.get("github_pr_id")
        pr_info.url = metadata.get("github_pr_url")
        pr_info.title = metadata.get("github_pr_title")
        pr_info.repository = metadata.get("github_repository")
    else:
        repo = hosting_repo()
        pr, message = recce_pr_information(github_token=kwargs.get("github_token"))
        if kwargs.get("cloud") and message:
            print(message)

        if pr:
            pr_info.repository = repo
            pr_info.id = pr.number
            pr_info.title = pr.title
            pr_info.url = pr.html_url
            pr_info.base_branch = pr.base.ref
            pr_info.branch = pr.head.ref

    # fetch from cli arguments
    if pr_info.url is None and "github_pull_request_url" in kwargs:
        pr_info.url = kwargs.get("github_pull_request_url")

    if pr_info.branch is None:
        pr_info.branch = kwargs.get("git_current_branch")

    if pr_info.base_branch is None:
        pr_info.base_branch = kwargs.get("git_base_branch")

    # fetch from env
    if pr_info.url is None:
        pr_info.url = os.getenv("RECCE_PR_URL")

    return pr_info


def fetch_pr_metadata_from_event_path() -> Optional[dict]:
    """
    If recce run is running in a GitHub Action, this function will return the pull request metadata.

    Example:
    {
        "github_pr_id": 1,
        "github_pr_url": "https://github.com/xyz/abc/pull/1,
        "github_pr_title": "Update README.md",
        "github_repository": "xyz/abc"
    }

    :return: dict
    """

    # get the event json from the path in GITHUB_EVENT_PATH
    event_path = os.getenv("GITHUB_EVENT_PATH")
    github_repository = os.getenv("GITHUB_REPOSITORY")
    if event_path:
        try:
            with open(event_path, "r", encoding="utf-8") as event_file:
                event_data = json.load(event_file)

            pr_id = event_data["number"]
            if event_data.get("pull_request"):
                pull_request_data = event_data["pull_request"]
                pr_url = pull_request_data["_links"]["html"]["href"]
                pr_api = pull_request_data["_links"]["self"]["href"]
                pr_title = _fetch_pr_title(pr_api)
                return dict(
                    github_pr_id=pr_id,
                    github_pr_url=pr_url,
                    github_pr_title=pr_title,
                    github_repository=github_repository,
                )
            else:
                print("Not a pull request event, skip.")
        except Exception as e:
            print("Cannot parse github action event", e)
    return None


def _fetch_pr_title(endpoint) -> Optional[str]:
    github_token = os.getenv("GITHUB_TOKEN")

    if github_token is None:
        return None

    try:
        headers = {"Authorization": f"Bearer {github_token}"}
        response = requests.get(endpoint, headers=headers)

        if response.status_code == 200:
            pull_request_data = response.json()
            return pull_request_data.get("title")
    except Exception as e:
        print("Cannot fetch PR title: ", e)

    return None



import logging
import os
import sys
import time
import traceback
import typing as t
from hashlib import sha256

from click import Context
from click.core import Command
from rich.console import Console
from rich.markup import escape

from recce import event, get_runner
from recce.core import load_context
from recce.exceptions import RecceException
from recce.git import current_branch, hosting_repo

console = Console()

_enable_traceback: bool = os.environ.get("RECCE_PRINT_TRACEBACK") == "1"
logger = logging.getLogger("uvicorn")


class TrackCommand(Command):
    def __init__(
        self,
        name: t.Optional[str],
        context_settings: t.Optional[t.Dict[str, t.Any]] = None,
        callback: t.Optional[t.Callable[..., t.Any]] = None,
        params: t.Any = None,
        help: t.Optional[str] = None,
        epilog: t.Optional[str] = None,
        short_help: t.Optional[str] = None,
        options_metavar: t.Optional[str] = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        beta: bool = False,
    ) -> None:
        super(TrackCommand, self).__init__(
            name,
            context_settings,
            callback,
            params,
            help,
            epilog,
            short_help,
            options_metavar,
            add_help_option,
            no_args_is_help,
            hidden,
            deprecated,
        )

    def _show_error_message(self, msg, params):
        if params.get("debug"):
            console.print_exception(show_locals=True)
        else:
            print(traceback.format_exc())
            # console.print('[bold red]Error:[/bold red] ', end='')
            # console.out(msg, highlight=False)

    def _show_hint_message(self, hint):
        console.print(f"[bold yellow]Hint[/bold yellow]:\n  {escape(hint)}")

    def invoke(self, ctx: Context) -> t.Any:
        status = False
        start_time = time.time()
        reason = "error"
        event.set_exception_tag("command", ctx.command.name)
        event.log_codespaces_events(ctx.command.name)

        try:
            ret = super(TrackCommand, self).invoke(ctx)
            if ret is None or ret == 0:
                status = True
                reason = "ok"
            else:
                reason = "error"
                sys.exit(ret)
            return ret
        except RecceException as e:
            logger.debug(traceback.format_exc())
            console.log("[Error] " + str(e))
            reason = "error"
            sys.exit(1)
        except SystemExit as e:
            reason = "error"
            raise e
        except KeyboardInterrupt as e:
            reason = "aborted"
            raise e
        except Exception as e:
            self._show_error_message(str(e), ctx.params)
            event.capture_exception(e)
            reason = "fatal"
            event.flush_exceptions()
            sys.exit(1)
        finally:
            end_time = time.time()
            runner = get_runner()
            repo = hosting_repo()
            branch = current_branch()
            command = ctx.command.name
            duration = end_time - start_time
            target_path = ctx.params.get("target_path", None)
            target_base_path = ctx.params.get("target_base_path", None)
            props = dict(
                command=command,
                status=status,
                reason=reason,
                duration=duration,
                cloud=ctx.params.get("cloud", False),
                review=ctx.params.get("review", False),
                debug=ctx.params.get("debug", False),
            )

            if runner is not None:
                props["runner_type"] = runner

            if repo is not None:
                props["repository"] = sha256(repo.encode()).hexdigest()

            if branch is not None:
                props["branch"] = sha256(branch.encode()).hexdigest()

            if target_path is not None:
                props["target_path"] = sha256(target_path.encode()).hexdigest()

            if target_base_path is not None:
                props["target_base_path"] = sha256(target_base_path.encode()).hexdigest()

            try:
                recce_context = load_context()
            except Exception:
                # it's not a ready-for-use project
                recce_context = None

            if recce_context is not None:
                if recce_context.adapter_type == "dbt":
                    props["adapter_type"] = "DBT"
                    # Add dbt warehouse type only for dbt adapter
                    try:
                        from recce.adapter.dbt_adapter import DbtAdapter

                        dbt_adapter: DbtAdapter = recce_context.adapter
                        warehouse_type = dbt_adapter.adapter.type()
                        props["warehouse_type"] = warehouse_type
                    except Exception:
                        # If we can't get the warehouse type, skip it
                        pass
                elif recce_context.adapter_type == "sqlmesh":
                    props["adapter_type"] = "SQLMesh"

            event.log_event(props, "command", params=ctx.params)
            event.flush_events()



import typing as t
from dataclasses import dataclass
from typing import Dict, Optional, Type

import pandas as pd
from sqlglot import Expression, parse_one, select
from sqlglot.optimizer import traverse_scope
from sqlmesh.core.context import Context as SqlmeshContext
from sqlmesh.core.environment import Environment
from sqlmesh.core.state_sync import StateReader

from recce.adapter.base import BaseAdapter
from recce.models import RunType
from recce.tasks import QueryDiffTask, QueryTask, RowCountDiffTask, Task

sqlmesh_supported_registry: Dict[RunType, Type[Task]] = {
    RunType.QUERY: QueryTask,
    RunType.QUERY_DIFF: QueryDiffTask,
    RunType.ROW_COUNT_DIFF: RowCountDiffTask,
}


@dataclass
class SqlmeshAdapter(BaseAdapter):
    context: SqlmeshContext
    base_env: Environment
    curr_env: Environment

    def support_tasks(self):
        return {run_type.value: True for run_type in sqlmesh_supported_registry}

    def get_lineage(self, base: Optional[bool] = False):
        state_reader: StateReader = self.context.state_reader
        env = self.base_env if base else self.curr_env
        nodes = {}
        parent_map = {}

        for snapshot in state_reader.get_snapshots(env.snapshots, hydrate_seeds=True).values():

            if snapshot.node_type.lower() != "model":
                continue

            model = snapshot.model
            if not model:
                continue

            node = {
                "unique_id": model.name,
                "name": model.name,
                "resource_type": snapshot.node_type.lower(),
                "checksum": {"checksum": snapshot.fingerprint.data_hash},
            }

            columns = {}
            for column, type in model.columns_to_types.items():
                columns[column] = {"name": column, "type": str(type)}
            node["columns"] = columns

            nodes[snapshot.name] = node
            parents = [snapshotId.name for snapshotId in snapshot.parents]
            parent_map[snapshot.name] = parents

        manifest_metadata = {}
        catalog_metadata = {}

        return dict(
            parent_map=parent_map,
            nodes=nodes,
            manifest_metadata=manifest_metadata,
            catalog_metadata=catalog_metadata,
        )

    def get_model(self, model_id: str, base=False):
        raise NotImplementedError()

    def get_node_name_by_id(self, unique_id):
        model = self.context.models.get(unique_id)
        if model is None:
            return None
        return model.name

    @classmethod
    def load(cls, **kwargs):
        sqlmesh_envs = kwargs.get("sqlmesh_envs")
        if sqlmesh_envs is None:
            raise Exception("'--sqlmesh-envs SOURCE:TARGET' is required")

        envs = sqlmesh_envs.split(":")
        if len(envs) != 2:
            raise Exception('sqlmesh_envs must be in the format of "SOURCE:TARGET"')

        sqlmesh_config = kwargs.get("sqlmesh_config", None)
        context = SqlmeshContext(config=sqlmesh_config)
        base_env = context.state_reader.get_environment(envs[0])
        curr_env = context.state_reader.get_environment(envs[1])
        if base_env is None:
            raise Exception(f"Source environment {envs[0]} not found")
        if curr_env is None:
            raise Exception(f"Target environment {envs[1]} not found")

        return cls(context=context, base_env=base_env, curr_env=curr_env)

    def replace_virtual_tables(self, sql: t.Union[Expression, str], base: bool = None) -> Expression:
        """
        Replace virtual tables based on the env name.

        Args:
            sql: SQL expression to replace virtual tables
            base: True: replace virtual tables with base env, False: replace virtual tables with current env, None: no replacement
        """
        if isinstance(sql, str):
            expression = parse_one(sql, dialect=self.context.default_dialect)
        else:
            expression = sql

        if base is not None:
            env = self.base_env if base else self.curr_env
            if env.name != "prod":
                model_names = [model.name for model in self.context.models.values()]
                for scope in traverse_scope(expression):
                    for table in scope.tables:
                        if f"{table.db}.{table.name}" in model_names:
                            table.args["db"] = f"{table.args['db']}__{env.name}"

        return expression

    def fetchdf_with_limit(
        self, sql: t.Union[Expression, str], base: Optional[bool] = None, limit: Optional[int] = None
    ) -> (pd.DataFrame, bool):
        expression = self.replace_virtual_tables(sql, base=base)
        if limit:
            expression = (
                select("*").from_("__QUERY").with_("__QUERY", as_=expression).limit(limit + 1 if limit else None)
            )
        df = self.context.fetchdf(expression)
        if limit and len(df) > limit:
            df = df.head(limit)
            more = True
        else:
            more = False
        return df, more



from typing import List, Literal, Optional, Union

from pydantic import BaseModel

from recce.models import Check
from recce.tasks.core import CheckValidator, TaskResultDiffer


class SchemaDiffResultDiffer:
    related_node_ids: List[str] = None

    def __init__(self, check, base_lineage, curr_lineage):
        self.check = check
        self.related_node_ids = self._get_related_node_ids()
        self.changes = self._check_result_changed_fn(base_lineage, curr_lineage)
        self.changed_nodes = self._get_changed_nodes()

    def _get_related_node_ids(self) -> Union[List[str], None]:
        params = self.check.params
        if params.get("node_id"):
            return params.get("node_id") if isinstance(params.get("node_id"), list) else [params.get("node_id")]
        else:
            return TaskResultDiffer.get_node_ids_by_selector(
                select=params.get("select"),
                exclude=params.get("exclude"),
                packages=params.get("packages"),
                view_mode=params.get("view_mode"),
            )

    def _check_result_changed_fn(self, base_lineage, curr_lineage):
        base = {}
        current = {}
        base_nodes = base_lineage.get("nodes", {})
        curr_nodes = curr_lineage.get("nodes", {})
        for node_id in self.related_node_ids:
            node = curr_nodes.get(node_id) or base_nodes.get(node_id)
            if not node:
                continue

            node_name = node.get("name")
            base[node_name] = base_nodes.get(node_id, {}).get("columns", {})
            current[node_name] = curr_nodes.get(node_id, {}).get("columns", {})

        return TaskResultDiffer.diff(base, current)

    def _get_changed_nodes(self) -> Union[List[str], None]:
        if self.changes:
            # Both affected_root_keys of deepdiff v7 (OrderedSet) and v8 (SetOrdered) are iterable
            # Convert to list directly
            return list(self.changes.affected_root_keys)
        return None


class SchemaDiffParams(BaseModel):
    node_id: Optional[Union[str, List[str]]] = None
    select: Optional[str] = None
    exclude: Optional[str] = None
    packages: Optional[list[str]] = None
    view_mode: Optional[Literal["all", "changed_models"]] = None


class SchemaDiffCheckValidator(CheckValidator):
    def validate_check(self, check: Check):
        SchemaDiffParams(**check.params)
        SchemaDiffParams(**check.view_options)



from collections import OrderedDict
from typing import Any


class LRUCache(object):
    def __init__(self, capacity: int = 128):
        self.capacity = capacity
        self.cache = OrderedDict()

    def get(self, key) -> Any:
        if key in self.cache:
            self.cache.move_to_end(key)
            return self.cache[key]
        else:
            return None

    def put(self, key, value):
        if key in self.cache:
            self.cache.move_to_end(key)
        elif len(self.cache) >= self.capacity:
            self.cache.popitem(last=False)
        self.cache[key] = value

    def clear(self):
        self.cache.clear()



"""Recce Cloud - Lightweight CLI for Recce Cloud operations."""

import os


def get_version():
    """Get version from main recce VERSION file."""
    # Reference the VERSION file from main recce package
    version_file = os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "recce", "VERSION"))
    with open(version_file) as fh:
        version = fh.read().strip()
        return version


__version__ = get_version()
